# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeHealthMonitorManagerFromDict(base_mgr.Nsxv3LoadbalancerBaseManager):
    def _build_monitor_args(self, hm):
        if hm['type'] in lb_const.NSXV3_MONITOR_MAP:
            monitor_type = lb_const.NSXV3_MONITOR_MAP.get(hm['type'])
        else:
            msg = (_('Cannot create health monitor %(monitor)s with '
                     'type %(type)s') % {'monitor': hm['id'],
                                         'type': hm['type']})
            raise n_exc.InvalidInput(error_message=msg)
        body = {'resource_type': monitor_type,
                'interval': hm['delay'],
                'fall_count': hm['max_retries'],
                'timeout': hm['timeout']}
        if hm['type'] in [lb_const.LB_HEALTH_MONITOR_HTTP,
                          lb_const.LB_HEALTH_MONITOR_HTTPS]:
            if hm['http_method']:
                body['request_method'] = hm['http_method']
            if hm['url_path']:
                body['request_url'] = hm['url_path']
            if hm['expected_codes']:
                if self.core_plugin.nsxlib.feature_supported(
                    nsx_constants.FEATURE_LB_HM_RESPONSE_CODES):
                    codes = hm['expected_codes'].split(",")
                    body['response_status_codes'] = [
                        int(code) for code in codes]
                else:
                    LOG.warning("Ignoring HM expected_codes as the NSX does "
                                "not support it")
        return body

    def create(self, context, hm, completor):
        lb_id = hm['pool']['loadbalancer_id']
        pool_id = hm['pool']['id']
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        monitor_client = self.core_plugin.nsxlib.load_balancer.monitor
        monitor_name = utils.get_name_and_uuid(hm['name'] or 'monitor',
                                               hm['id'])
        tags = lb_utils.get_tags(self.core_plugin, hm['id'],
                                 lb_const.LB_HM_TYPE,
                                 hm['tenant_id'], context.project_name)
        monitor_body = self._build_monitor_args(hm)

        try:
            lb_monitor = monitor_client.create(
                display_name=monitor_name, tags=tags, **monitor_body)
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)

        binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool_id)
        if binding:
            lb_pool_id = binding['lb_pool_id']
            try:
                pool_client.add_monitor_to_pool(lb_pool_id,
                                                lb_monitor['id'])
            except nsxlib_exc.ManagerError:
                completor(success=False)
                msg = _('Failed to attach monitor %(monitor)s to pool '
                        '%(pool)s') % {'monitor': lb_monitor['id'],
                                       'pool': lb_pool_id}
                raise n_exc.BadRequest(resource='lbaas-hm', msg=msg)
            nsx_db.add_nsx_lbaas_monitor_binding(
                context.session, lb_id, pool_id, hm['id'], lb_monitor['id'],
                lb_pool_id)
        else:
            completor(success=False)
            msg = _('Failed to attach monitor %(monitor)s to pool '
                    '%(pool)s: NSX pool was not found on the DB') % {
                    'monitor': hm['id'],
                    'pool': pool_id}
            raise n_exc.BadRequest(resource='lbaas-hm', msg=msg)

        completor(success=True)

    def update(self, context, old_hm, new_hm, completor):
        lb_id = new_hm['pool']['loadbalancer_id']
        pool_id = new_hm['pool']['id']
        monitor_client = self.core_plugin.nsxlib.load_balancer.monitor
        binding = nsx_db.get_nsx_lbaas_monitor_binding(
            context.session, lb_id, pool_id, new_hm['id'])
        if binding:
            lb_monitor_id = binding['lb_monitor_id']
            monitor_body = self._build_monitor_args(new_hm)
            monitor_name = utils.get_name_and_uuid(new_hm['name'] or 'monitor',
                                                   new_hm['id'])
            monitor_client.update(lb_monitor_id, display_name=monitor_name,
                                  **monitor_body)
        else:
            completor(success=False)
            msg = _('Failed to update monitor %(monitor)s: NSX monitor was '
                    'not found in DB') % {'monitor': new_hm['id'],
                                          'pool': pool_id}
            raise n_exc.BadRequest(resource='lbaas-hm', msg=msg)

        completor(success=True)

    def delete(self, context, hm, completor):
        lb_id = hm['pool']['loadbalancer_id']
        pool_id = hm['pool']['id']
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        monitor_client = self.core_plugin.nsxlib.load_balancer.monitor

        binding = nsx_db.get_nsx_lbaas_monitor_binding(
            context.session, lb_id, pool_id, hm['id'])
        if binding:
            lb_monitor_id = binding['lb_monitor_id']
            lb_pool_id = binding['lb_pool_id']
            try:
                pool_client.remove_monitor_from_pool(lb_pool_id,
                                                     lb_monitor_id)
            except nsxlib_exc.ResourceNotFound:
                pass
            except nsxlib_exc.ManagerError as exc:
                completor(success=False)
                msg = _('Failed to remove monitor %(monitor)s from pool '
                        '%(pool)s with exception from nsx %(exc)s)') % {
                    'monitor': lb_monitor_id,
                    'pool': lb_pool_id,
                    'exc': exc}
                raise n_exc.BadRequest(resource='lbaas-hm', msg=msg)
            try:
                monitor_client.delete(lb_monitor_id)
            except nsxlib_exc.ResourceNotFound:
                pass
            except nsxlib_exc.ManagerError as exc:
                completor(success=False)
                msg = _('Failed to delete monitor %(monitor)s from '
                        'backend with exception %(exc)s') % {
                          'monitor': lb_monitor_id,
                          'exc': exc}
                raise n_exc.BadRequest(resource='lbaas-hm', msg=msg)

            nsx_db.delete_nsx_lbaas_monitor_binding(context.session, lb_id,
                                                    pool_id, hm['id'])
        else:
            # Do not fail a delete action
            pass

        completor(success=True)

    def delete_cascade(self, context, hm, completor):
        self.delete(context, hm, completor)



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api import extensions
from neutron_lib import constants
from neutron_lib import exceptions as nexception

from vmware_nsx._i18n import _


ALIAS = 'provider-security-group'
PROVIDER = 'provider'
PROVIDER_SECURITYGROUPS = 'provider_security_groups'

EXTENDED_ATTRIBUTES_2_0 = {
    'security_groups': {
        PROVIDER: {
            'allow_post': True,
            'allow_put': False,
            'convert_to': converters.convert_to_boolean,
            'default': False,
            'enforce_policy': True,
            'is_visible': True}
    },
    'ports': {PROVIDER_SECURITYGROUPS: {
        'allow_post': True,
        'allow_put': True,
        'is_visible': True,
        'convert_to': converters.convert_none_to_empty_list,
        'validate': {'type:uuid_list': None},
        'default': constants.ATTR_NOT_SPECIFIED}
    }
}


NUM_PROVIDER_SGS_ON_PORT = 1


class SecurityGroupNotProvider(nexception.InvalidInput):
    message = _("Security group %(id)s is not a provider security group.")


class SecurityGroupIsProvider(nexception.InvalidInput):
    message = _("Security group %(id)s is a provider security group and "
                "cannot be specified via the security group field.")


class DefaultSecurityGroupIsNotProvider(nexception.InvalidInput):
    message = _("Can't create default security-group as a provider "
                "security-group.")


class ProviderSecurityGroupEditNotAdmin(nexception.NotAuthorized):
    message = _("Security group %(id)s is a provider security group and "
                "requires an admin to modify it.")


class Providersecuritygroup(extensions.ExtensionDescriptor):
    """Provider security-group extension."""

    @classmethod
    def get_name(cls):
        return "Provider security group"

    @classmethod
    def get_alias(cls):
        return ALIAS

    @classmethod
    def get_description(cls):
        return "Admin controlled security groups with blocking rules."

    @classmethod
    def get_updated(cls):
        return "2016-07-13T10:00:00-00:00"

    def get_required_extensions(self):
        return ["security-group"]

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        return []

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2018 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxp.resources import utils as p_utils
from vmware_nsx.shell import resources as shell

from neutron_lib.callbacks import registry


LOG = logging.getLogger(__name__)

MIN_REALIZATION_INTERVAL = 1
MAX_REALIZATION_INTERVAL = 10


def set_system_parameters(resource, event, trigger, **kwargs):
    """Set interval that controls realization and purge frequency

    This setting is affecting NSX Policy Manager appliance.
    """
    if kwargs.get('property'):
        properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
        interval = properties.get('realization_interval')

        if interval:
            int_interval = int(interval)
            if int_interval not in range(MIN_REALIZATION_INTERVAL,
                                         MAX_REALIZATION_INTERVAL + 1):
                LOG.info("Realization interval should be in range %d-%d",
                         MIN_REALIZATION_INTERVAL, MAX_REALIZATION_INTERVAL)
                return

            nsxpolicy = p_utils.get_connected_nsxpolicy()
            try:
                nsxpolicy.set_realization_interval(int_interval)
            except Exception as ex:
                LOG.error("Failed to apply intent realization interval to "
                          "policy appliance - %s", ex)

            LOG.info("Intent realization interval set to %s min" % interval)


registry.subscribe(set_system_parameters,
                   constants.SYSTEM,
                   shell.Operations.SET.value)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils as p_utils
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsx.services.lbaas.octavia import constants as oct_const
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3.policy import utils as lib_p_utils

LOG = logging.getLogger(__name__)


class EdgeLoadBalancerManagerFromDict(base_mgr.NsxpLoadbalancerBaseManager):

    def _validate_lb_network(self, context, lb):
        router_id = lb_utils.get_router_from_network(
            context, self.core_plugin, lb['vip_subnet_id'])

        return router_id

    def _get_info_from_fip(self, context, fip):
        filters = {'floating_ip_address': [fip]}
        floating_ips = self.core_plugin.get_floatingips(context,
                                                        filters=filters)
        if floating_ips:
            return (floating_ips[0]['fixed_ip_address'],
                    floating_ips[0]['router_id'])
        else:
            msg = (_('Member IP %(fip)s is an external IP, and is expected to '
                     'be a floating IP') % {'fip': fip})
            raise n_exc.BadRequest(resource='lbaas-vip', msg=msg)

    def create(self, context, lb, completor):
        lb_id = lb['id']

        network = lb_utils.get_network_from_subnet(
            context, self.core_plugin, lb['vip_subnet_id'])

        router_id = self._validate_lb_network(context, lb)
        if not router_id and network and not network.get('router:external'):
            completor(success=False)
            msg = (_('Cannot create a loadbalancer %(lb_id)s on subnet. '
                     '%(subnet)s is neither public nor connected to the LB '
                     'router') %
                   {'lb_id': lb_id, 'subnet': lb['vip_subnet_id']})
            raise n_exc.BadRequest(resource='lbaas-subnet', msg=msg)

        service_client = self.core_plugin.nsxpolicy.load_balancer.lb_service
        lb_service = None
        if router_id:
            # Check if a service was already created for this router
            lb_service = p_utils.get_router_nsx_lb_service(
                self.core_plugin.nsxpolicy, router_id)

            if lb_service:
                # Add the new LB to the service by adding the tag.
                # At the same time verify maximum number of tags
                try:
                    with p_utils.get_lb_rtr_lock(router_id):
                        service_client.update_customized(
                            lb_service['id'],
                            p_utils.add_service_tag_callback(lb['id']))
                except n_exc.BadRequest:
                    # This will fail if there are too many loadbalancers
                    completor(success=False)
                    msg = (_('Cannot create a loadbalancer %(lb_id)s on '
                             'router %(router)s, as it already has too many '
                             'loadbalancers') %
                           {'lb_id': lb_id, 'router': router_id})
                    raise n_exc.BadRequest(resource='lbaas-router', msg=msg)
                except Exception as e:
                    with excutils.save_and_reraise_exception():
                        completor(success=False)
                        LOG.error('Failed to create loadbalancer %(lb)s for '
                                  'lb with exception %(e)s',
                                  {'lb': lb['id'], 'e': e})
            else:
                # Create the Tier1 service router if it does not exist
                if not self.core_plugin.service_router_has_services(
                    context.elevated(), router_id):
                    self.core_plugin.create_service_router(context, router_id)

        if not lb_service:
            lb_name = p_utils.get_service_lb_name(lb, router_id)
            tags = p_utils.get_tags(self.core_plugin,
                                    router_id if router_id else '',
                                    lb_const.LR_ROUTER_TYPE,
                                    lb['tenant_id'], context.project_name)
            tags.append(p_utils.get_service_lb_tag(lb['id']))

            lb_size = lb_utils.get_lb_flavor_size(self.flavor_plugin, context,
                                                  lb.get('flavor_id'))

            try:
                if network and network.get('router:external'):
                    connectivity_path = None
                else:
                    tier1_srv = self.core_plugin.nsxpolicy.tier1
                    connectivity_path = tier1_srv.get_path(router_id)
                if connectivity_path:
                    with p_utils.get_lb_rtr_lock(router_id):
                        service_client.create_or_overwrite(
                            lb_name, lb_service_id=lb['id'],
                            description=lb['description'],
                            tags=tags, size=lb_size,
                            connectivity_path=connectivity_path)
                else:
                    # no lock
                    service_client.create_or_overwrite(
                        lb_name, lb_service_id=lb['id'],
                        description=lb['description'],
                        tags=tags, size=lb_size,
                        connectivity_path=connectivity_path)

                # Add rule to advertise external vips
                if router_id:
                    p_utils.update_router_lb_vip_advertisement(
                        context, self.core_plugin, router_id)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    completor(success=False)
                    LOG.error('Failed to create loadbalancer %(lb)s for lb '
                              'with exception %(e)s', {'lb': lb['id'], 'e': e})

        # Make sure the vip port is marked with a device owner
        port = self.core_plugin.get_port(
            context.elevated(), lb['vip_port_id'])
        if not port.get('device_owner'):
            self.core_plugin.update_port(
                context.elevated(), lb['vip_port_id'],
                {'port': {'device_id': oct_const.DEVICE_ID_PREFIX + lb['id'],
                          'device_owner': lb_const.VMWARE_LB_VIP_OWNER}})
        completor(success=True)

    def update(self, context, old_lb, new_lb, completor):
        completor(success=True)

    def delete(self, context, lb, completor):
        router_id = None
        try:
            router_id = lb_utils.get_router_from_network(
                context, self.core_plugin, lb['vip_subnet_id'])
        except n_exc.SubnetNotFound:
            LOG.warning("VIP subnet %s not found while deleting "
                        "loadbalancer %s", lb['vip_subnet_id'], lb['id'])
        service_client = self.core_plugin.nsxpolicy.load_balancer.lb_service
        service = p_utils.get_lb_nsx_lb_service(
            self.core_plugin.nsxpolicy, lb['id'])
        if service:
            lb_service_id = service['id']
            if not router_id:
                router_id = lib_p_utils.path_to_id(
                    service.get('connectivity_path', ''))

            if router_id:
                with p_utils.get_lb_rtr_lock(router_id):
                    # check if this is the last lb for this router and update
                    # tags / delete service
                    try:
                        service_client.update_customized(
                            lb_service_id,
                            p_utils.remove_service_tag_callback(lb['id']))
                    except n_exc.BadRequest:
                        # This is the last Lb and service should be deleted
                        try:
                            service_client.delete(lb_service_id)
                        except Exception as e:
                            with excutils.save_and_reraise_exception():
                                completor(success=False)
                                LOG.error('Failed to delete loadbalancer '
                                          '%(lb)s for lb with error %(err)s',
                                          {'lb': lb['id'], 'err': e})
                        # Remove the service router if no more services
                        if not self.core_plugin.service_router_has_services(
                                context.elevated(), router_id):
                            try:
                                self.core_plugin.delete_service_router(
                                    router_id)
                            except Exception as e:
                                with excutils.save_and_reraise_exception():
                                    completor(success=False)
                                    LOG.error('Failed to delete service '
                                              'router upon deletion '
                                              'of loadbalancer %(lb)s with '
                                              'error %(err)s',
                                              {'lb': lb['id'], 'err': e})
            else:
                # LB without router (meaning external vip and no members)
                try:
                    service_client.delete(lb_service_id)
                except Exception as e:
                    with excutils.save_and_reraise_exception():
                        completor(success=False)
                        LOG.error('Failed to delete loadbalancer %(lb)s for '
                                  'lb with error %(err)s',
                                  {'lb': lb['id'], 'err': e})

        # Make sure the vip port is not marked with a vmware device owner
        try:
            port = self.core_plugin.get_port(
                context.elevated(), lb['vip_port_id'])
            if port.get('device_owner') == lb_const.VMWARE_LB_VIP_OWNER:
                self.core_plugin.update_port(
                    context.elevated(), lb['vip_port_id'],
                    {'port': {'device_id': '',
                              'device_owner': ''}})
        except n_exc.PortNotFound:
            # Only log the error and continue anyway
            LOG.warning("VIP port %s not found while deleting loadbalancer %s",
                        lb['vip_port_id'], lb['id'])
        except Exception as e:
            # Just log the error as all other resources were deleted
            LOG.error("Failed to update neutron port %s devices upon "
                      "loadbalancer deletion: %s", lb['vip_port_id'], e)

        completor(success=True)

    def delete_cascade(self, context, lb, completor):
        """Delete all backend and DB resources of this loadbalancer"""
        self.delete(context, lb, completor)

    def refresh(self, context, lb):
        # TODO(kobis): implement
        pass

    def _get_lb_virtual_servers(self, context, lb):
        # Get all virtual servers that belong to this loadbalancer
        vs_list = [vs['id'] for vs in lb['listeners']]
        return vs_list

    def stats(self, context, lb):
        # Since multiple LBaaS loadbalancer can share the same LB service,
        # get the corresponding virtual servers' stats instead of LB service.
        stats = {'active_connections': 0,
                 'bytes_in': 0,
                 'bytes_out': 0,
                 'total_connections': 0}

        service_client = self.core_plugin.nsxpolicy.load_balancer.lb_service
        vs_list = self._get_lb_virtual_servers(context, lb)
        try:
            rsp = service_client.get_statistics(lb['id'])
            for result in rsp.get('results', []):
                for vs in result.get('virtual_servers', []):
                    # Skip the virtual server that doesn't belong
                    # to this loadbalancer
                    vs_id = lib_p_utils.path_to_id(vs['virtual_server_path'])
                    if vs_id not in vs_list:
                        continue
                    vs_stats = vs.get('statistics', {})
                    for stat in lb_const.LB_STATS_MAP:
                        lb_stat = lb_const.LB_STATS_MAP[stat]
                        stats[stat] += vs_stats.get(lb_stat, 0)

        except nsxlib_exc.ManagerError:
            msg = _('Failed to retrieve stats from LB service '
                    'for loadbalancer %(lb)s') % {'lb': lb['id']}
            raise n_exc.BadRequest(resource='lbaas-lb', msg=msg)
        return stats

    def get_operating_status(self, context, id, with_members=False):
        service_client = self.core_plugin.nsxpolicy.load_balancer.lb_service
        try:
            service_status = service_client.get_status(id)
            if not isinstance(service_status, dict):
                service_status = {}

        except nsxlib_exc.ManagerError:
            LOG.warning("LB service %(lbs)s is not found",
                        {'lbs': id})
            return {}

        # get the loadbalancer status from the LB service
        lb_status_results = service_status.get('results')
        lb_status = lb_const.ONLINE
        if lb_status_results:
            result = lb_status_results[0]
            lb_status = _get_octavia_lb_status(result)

        statuses = {lb_const.LOADBALANCERS: [{'id': id, 'status': lb_status}],
                    lb_const.LISTENERS: [],
                    lb_const.POOLS: [],
                    lb_const.MEMBERS: []}

        # TODO(asarfaty): Go over all VS of this loadbalancer by tags
        # to add the listeners statuses from the virtual servers statuses
        return statuses


def _nsx_status_to_lb_status(nsx_status):
    if not nsx_status:
        # default fallback
        return lb_const.ONLINE

    # Statuses that are considered ONLINE:
    if nsx_status.upper() in ['UP', 'UNKNOWN', 'PARTIALLY_UP',
                              'NO_STANDBY']:
        return lb_const.ONLINE
    # Statuses that are considered OFFLINE:
    if nsx_status.upper() in ['PRIMARY_DOWN', 'DETACHED', 'DOWN', 'ERROR']:
        return lb_const.OFFLINE
    if nsx_status.upper() == 'DISABLED':
        return lb_const.DISABLED

    # default fallback
    LOG.debug("NSX LB status %s - interpreted as ONLINE", nsx_status)
    return lb_const.ONLINE


def _get_octavia_lb_status(result):
    if result.get('service_status'):
        # Use backend service_status
        lb_status = _nsx_status_to_lb_status(
            result['service_status'])
    elif result.get('alarm'):
        # No status, but has alarms -> ERROR
        lb_status = lb_const.OFFLINE
    else:
        # Unknown - assume it is ok
        lb_status = lb_const.ONLINE
    return lb_status


def status_getter(context, core_plugin):
    nsxlib_lb = core_plugin.nsxpolicy.load_balancer
    lb_client = nsxlib_lb.lb_service
    lbs = lb_client.list()
    lb_statuses = []
    lsn_statuses = []
    pool_statuses = []
    member_statuses = []
    for lb in lbs:
        try:
            service_status = lb_client.get_status(lb['id'])
            if not isinstance(service_status, dict):
                service_status = {}
        except nsxlib_exc.ManagerError:
            LOG.warning("LB service %(lbs)s is not found",
                        {'lbs': id})
            service_status = {}
        lb_status_results = service_status.get('results')
        if lb_status_results:
            result = lb_status_results[0]
            lb_operating_status = _get_octavia_lb_status(result)
            for vs_status in result.get('virtual_servers', []):
                vs_id = lib_p_utils.path_to_id(
                    vs_status['virtual_server_path'])
                lsn_statuses.append({
                    'id': vs_id,
                    'operating_status': _nsx_status_to_lb_status(
                        vs_status['status'])})
            for pool_status in result.get('pools', []):
                pool_id = lib_p_utils.path_to_id(pool_status['pool_path'])
                pool_statuses.append({
                    'id': pool_id,
                    'operating_status': _nsx_status_to_lb_status(
                        pool_status['status'])})
                for member in pool_status.get('members', []):
                    member_statuses.append({
                        'pool_id': pool_id,
                        'member_ip': member.get('ip_address'),
                        'operating_status': _nsx_status_to_lb_status(
                            member['status'])})

        else:
            lb_operating_status = lb_const.OFFLINE

        for tag in lb['tags']:
            if tag['scope'] == 'loadbalancer_id':
                lb_statuses.append(
                    {'id': tag['tag'],
                     'operating_status': lb_operating_status})

    return {lb_const.LOADBALANCERS: lb_statuses,
            lb_const.LISTENERS: lsn_statuses,
            lb_const.POOLS: pool_statuses,
            lb_const.MEMBERS: member_statuses}



# Copyright 2020 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import registry
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging

from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxp.resources import utils as p_utils
from vmware_nsx.shell import resources as shell

LOG = logging.getLogger(__name__)


@admin_utils.output_header
def update_lb_service_tags(resource, event, trigger, **kwargs):
    """Update the LB id tag on existing LB services"""
    nsxpolicy = p_utils.get_connected_nsxpolicy()
    service_client = nsxpolicy.load_balancer.lb_service
    services = service_client.list()
    n_updated = 0
    for lb_service in services:
        # First make sure it i a neutron service
        is_neutron = False
        for tag in lb_service.get('tags', []):
            if tag['scope'] == 'os-api-version':
                is_neutron = True
                break
        if is_neutron:
            # Add a tag with the id of this resource as the first Lb
            # creates the service with its id
            try:
                service_client.update_customized(
                    lb_service['id'],
                    lb_utils.add_service_tag_callback(lb_service['id'],
                                                      only_first=True))
            except n_exc.BadRequest:
                LOG.warning("Lb service %s already has a loadbalancer tag",
                            lb_service['id'])
            else:
                n_updated = n_updated + 1

    LOG.info("Done updating %s Lb services.", n_updated)


registry.subscribe(update_lb_service_tags,
                   constants.LB_SERVICES,
                   shell.Operations.NSX_UPDATE_TAGS.value)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import xml.etree.ElementTree as et

from neutron_lbaas.db.loadbalancer import models as nlbaas_v2
from neutron_lib.callbacks import registry

from vmware_nsx.common import locking
from vmware_nsx.plugins.nsx_v.vshield import vcns as nsxv_api
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common.utils import output_header
from vmware_nsx.shell.admin.plugins.nsxv.resources import utils as utils
from vmware_nsx.shell.resources import Operations

LBAAS_FW_SECTION_NAME = 'LBaaS FW Rules'
LOG = logging.getLogger(__name__)


@output_header
def sync_lbaas_dfw_rules(resource, event, trigger, **kwargs):
    vcns = utils.get_nsxv_client()
    with locking.LockManager.get_lock('lbaas-fw-section'):
        fw_section_id = vcns.get_section_id(LBAAS_FW_SECTION_NAME)
        if not fw_section_id:
            section = et.Element('section')
            section.attrib['name'] = LBAAS_FW_SECTION_NAME
            sect = vcns.create_section('ip', et.tostring(section))[1]
            fw_section_id = et.fromstring(sect).attrib['id']

        if not fw_section_id:
            LOG.error('No LBaaS FW Section id found')
            return

        neutron_db = utils.NeutronDbClient()
        pools = neutron_db.context.session.query(nlbaas_v2.PoolV2).all()
        pool_ids = [pool['id'] for pool in pools]

        section_uri = '%s/%s/%s' % (nsxv_api.FIREWALL_PREFIX,
                                    'layer3sections',
                                    fw_section_id)

        xml_section_data = vcns.get_section(section_uri)
        if xml_section_data:
            xml_section = xml_section_data[1]
        else:
            LOG.info('LBaaS XML section was not found!')
            return

        section = et.fromstring(xml_section)

        for rule in section.findall('.//rule'):
            if rule.find('name').text in pool_ids:
                LOG.info('Rule %s found and valid', rule.find('name').text)
            else:
                section.remove(rule)
                LOG.info('Rule %s is stale and removed',
                         rule.find('name').text)

        vcns.update_section(section_uri,
                            et.tostring(section, encoding="us-ascii"),
                            None)


registry.subscribe(sync_lbaas_dfw_rules,
                   constants.LBAAS,
                   Operations.NSX_UPDATE.value)



# Copyright 2017 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from neutron.extensions import securitygroup as ext_sg
from neutron_lib.callbacks import registry
from neutron_lib import context as n_context
from neutron_lib import exceptions

from vmware_nsx.api_replay import utils as replay_utils
from vmware_nsx.db import db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv.resources import utils as v_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import utils as v3_utils
from vmware_nsx.shell import resources as shell

LOG = logging.getLogger(__name__)
# list of supported objects to migrate in order of deletion (creation will be
# in the opposite order)
migrated_resources = ["floatingip", "router", "port", "subnet",
                      "network", "security_group"]
#TODO(asarfaty): add other resources of different service plugins like
#vpnaas, fwaas, lbaas, qos, subnetpool, etc


@admin_utils.output_header
def import_projects(resource, event, trigger, **kwargs):
    """Import existing openstack projects to the current plugin"""
    # TODO(asarfaty): get the projects list from keystone
    # get the plugin name from the user
    if not kwargs.get('property'):
        LOG.error("Need to specify plugin and project parameters")
        return
    else:
        properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
        plugin = properties.get('plugin')
        project = properties.get('project')
        if not plugin or not project:
            LOG.error("Need to specify plugin and project parameters")
            return
    if plugin not in projectpluginmap.VALID_TYPES:
        LOG.error("The supported plugins are %s", projectpluginmap.VALID_TYPES)
        return

    ctx = n_context.get_admin_context()
    curr_map = db.get_project_plugin_mapping(ctx.session, project)
    if not curr_map:
        db.add_project_plugin_mapping(ctx.session, project, plugin)
        LOG.info('Done.')
    else:
        if curr_map.plugin == plugin:
            LOG.info("%s is already the plugin of project %s", plugin, project)
        else:
            LOG.error("Project %s is mapped to plugin %s. Existing mapping "
                      "cannot be modified.", project, curr_map.plugin)


def get_resource_file_name(project_id, resource):
    return "%s_nsxv_%ss" % (project_id, resource)


def read_v_resources_to_files(context, project_id):
    """Read all relevant NSX-V resources from a specific project

    and write them into a json file
    """
    results = {}
    with v_utils.NsxVPluginWrapper() as plugin:
        filters = {'project_id': [project_id]}
        for resource in migrated_resources:
            filename = get_resource_file_name(project_id, resource)
            file = open(filename, 'w')
            get_objects = getattr(plugin, "get_%ss" % resource)
            objects = get_objects(context, filters=filters)

            # also add router gateway ports of the relevant routers
            # (don't have the project id)
            if resource == 'port':
                rtr_ids = [rtr['id'] for rtr in results['router']]
                gw_filters = {'device_owner': ['network:router_gateway'],
                              'device_id': rtr_ids}
                gw_ports = plugin.get_ports(context, filters=gw_filters,
                                            filter_project=False)
                # ignore metadata gw ports
                objects.extend([port for port in gw_ports
                                if not port['tenant_id']])

            file.write(jsonutils.dumps(objects, sort_keys=True, indent=4))
            file.close()
            results[resource] = objects

    return results


def read_v_resources_from_files(project_id):
    """Read all relevant NSX-V resources from a json file"""
    results = {}
    for resource in migrated_resources:
        filename = get_resource_file_name(project_id, resource)
        file = open(filename, 'r')
        results[resource] = jsonutils.loads(file.read())
        file.close()
    return results


def delete_router_routes_and_interfaces(context, plugin, router):
    if router.get('routes'):
        plugin.update_router(context, router['id'],
                             {'router': {'routes': []}})

    interfaces = plugin._get_router_interfaces(context, router['id'])
    for port in interfaces:
        plugin.remove_router_interface(context, router['id'],
                                       {'port_id': port['id']})


def delete_v_resources(context, objects):
    """Delete a list of objects from the V plugin"""
    with v_utils.NsxVPluginWrapper() as plugin:
        LOG.info(">>>>Deleting all NSX-V objects of the project.")
        for resource in migrated_resources:
            get_object = getattr(plugin, "get_%s" % resource)
            del_object = getattr(plugin, "delete_%s" % resource)
            for obj in objects[resource]:
                # verify that this object still exists
                try:
                    get_object(context, obj['id'])
                except exceptions.NotFound:
                    # prevent logger from logging this exception
                    sys.exc_clear()
                    continue

                try:
                    # handle special cases before delete
                    if resource == 'router':
                        delete_router_routes_and_interfaces(
                            context, plugin, obj)
                    elif resource == 'port':
                        if obj['device_owner'] == 'network:dhcp':
                            continue
                    # delete the objects from the NSX-V plugin
                    del_object(context, obj['id'])
                    LOG.info(">>Deleted %(resource)s %(name)s",
                             {'resource': resource,
                              'name': obj.get('name') or obj['id']})
                except Exception as e:
                    LOG.warning(">>Failed to delete %(resource)s %(name)s: "
                                "%(e)s",
                                {'resource': resource,
                                 'name': obj.get('name') or obj['id'], 'e': e})
    LOG.info(">>>>Done deleting all NSX-V objects.")


def get_router_by_id(objects, router_id):
    for rtr in objects.get('router', []):
        if rtr['id'] == router_id:
            return rtr


def create_t_resources(context, objects, ext_net):
    """Create a list of objects in the T plugin"""
    LOG.info(">>>>Creating all the objects of the project in NSX-T.")
    prepare = replay_utils.PrepareObjectForMigration()
    with v3_utils.NsxV3PluginWrapper() as plugin:
        # create the resource in the order opposite to the deletion
        # (but start with routers)
        ordered_resources = migrated_resources[::-1]
        ordered_resources.remove('router')
        ordered_resources = ['router'] + ordered_resources
        dhcp_subnets = []
        for resource in ordered_resources:
            total_num = len(objects[resource])
            LOG.info(">>>Creating %s %s%s.", total_num,
                     resource, 's' if total_num > 1 else '')
            get_object = getattr(plugin, "get_%s" % resource)
            create_object = getattr(plugin, "create_%s" % resource)
            # go over the objects of this resource
            for count, obj in enumerate(objects[resource], 1):
                # check if this object already exists
                try:
                    get_object(context, obj['id'])
                except exceptions.NotFound:
                    # prevent logger from logging this exception
                    sys.exc_clear()
                else:
                    # already exists (this will happen if we rerun from files,
                    # or if the deletion failed)
                    LOG.info(">>Skipping %(resource)s %(name)s %(count)s/"
                             "%(total)s as it was already created.",
                             {'resource': resource,
                              'name': obj.get('name') or obj['id'],
                              'count': count,
                              'total': total_num})
                    continue

                # fix object before creation using the api replay code
                orig_id = obj['id']
                prepare_object = getattr(prepare, "prepare_%s" % resource)
                # TODO(asarfaty): Add availability zones support too
                obj_data = prepare_object(obj, direct_call=True)
                enable_dhcp = False
                # special cases for different objects before create:
                if resource == 'subnet':
                    if obj_data['enable_dhcp']:
                        enable_dhcp = True
                        # disable dhcp for now, to avoid ip collisions
                        obj_data['enable_dhcp'] = False
                elif resource == 'security_group':
                    # security group rules should be added separately
                    sg_rules = obj_data.pop('security_group_rules')
                elif resource == 'floatingip':
                    # Create the floating IP on the T external network
                    obj_data['floating_network_id'] = ext_net
                    del obj_data['floating_ip_address']
                elif resource == 'port':
                    # remove the old subnet id field from ports fixed_ips dict
                    # since the subnet ids are changed
                    for fixed_ips in obj_data['fixed_ips']:
                        del fixed_ips['subnet_id']

                    if obj_data['device_owner'] == 'network:dhcp':
                        continue
                    if obj_data['device_owner'] == 'network:floatingip':
                        continue
                    if obj_data['device_owner'] == 'network:router_gateway':
                        # add a gateway on the new ext network for this router
                        router_id = obj_data['device_id']
                        # keep the original enable-snat value
                        router_data = get_router_by_id(objects, router_id)
                        enable_snat = router_data['external_gateway_info'].get(
                                'enable_snat', True)
                        rtr_body = {
                            "external_gateway_info":
                                {"network_id": ext_net,
                                 "enable_snat": enable_snat}}
                        try:
                            plugin.update_router(
                                context, router_id, {'router': rtr_body})
                            LOG.info(">>Uplinked router %(rtr)s to new "
                                     "external network %(net)s",
                                     {'rtr': router_id,
                                      'net': ext_net})

                        except Exception as e:
                            LOG.error(">>Failed to add router %(rtr)s "
                                      "gateway: %(e)s",
                                      {'rtr': router_id, 'e': e})
                            continue
                    if obj_data['device_owner'] == 'network:router_interface':
                        try:
                            # uplink router_interface ports by creating the
                            # port, and attaching it to the router
                            router_id = obj_data['device_id']
                            obj_data['device_owner'] = ""
                            obj_data['device_id'] = ""
                            created_port = plugin.create_port(
                                context,
                                {'port': obj_data})
                            LOG.info(">>Created interface port %(port)s, ip "
                                     "%(ip)s, mac %(mac)s)",
                                     {'port': created_port['id'],
                                      'ip': created_port['fixed_ips'][0][
                                            'ip_address'],
                                      'mac': created_port['mac_address']})
                            plugin.add_router_interface(
                                context,
                                router_id,
                                {'port_id': created_port['id']})
                            LOG.info(">>Uplinked router %(rtr)s to network "
                                     "%(net)s",
                                     {'rtr': router_id,
                                      'net': obj_data['network_id']})
                        except Exception as e:
                            LOG.error(">>Failed to add router %(rtr)s "
                                      "interface port: %(e)s",
                                      {'rtr': router_id, 'e': e})
                        continue

                # create the object on the NSX-T plugin
                try:
                    created_obj = create_object(context, {resource: obj_data})
                    LOG.info(">>Created %(resource)s %(name)s %(count)s/"
                             "%(total)s",
                             {'resource': resource, 'count': count,
                              'name': obj_data.get('name') or orig_id,
                              'total': total_num})
                except Exception as e:
                    # TODO(asarfaty): subnets ids are changed, so recreating a
                    # subnet will fail on overlapping ips.
                    LOG.error(">>Failed to create %(resource)s %(name)s: "
                              "%(e)s",
                              {'resource': resource, 'e': e,
                               'name': obj_data.get('name') or orig_id})
                    continue

                # special cases for different objects after create:
                if resource == 'security_group':
                    sg_id = obj_data.get('name') or obj_data['id']
                    for rule in sg_rules:
                        rule_data = prepare.prepare_security_group_rule(rule)
                        try:
                            plugin.create_security_group_rule(
                                context, {'security_group_rule': rule_data})
                        except ext_sg.SecurityGroupRuleExists:
                            # default rules were already created.
                            # prevent logger from logging this exception
                            sys.exc_clear()
                        except Exception as e:
                            LOG.error(
                                ">>Failed to create security group %(name)s "
                                "rules: %(e)s",
                                {'name': sg_id, 'e': e})
                elif resource == 'subnet':
                    if enable_dhcp:
                        dhcp_subnets.append(created_obj['id'])

        # Enable dhcp on all the relevant subnets (after creating all ports,
        # to maintain original IPs):
        if dhcp_subnets:
            for subnet_id in dhcp_subnets:
                try:
                    plugin.update_subnet(
                        context, subnet_id,
                        {'subnet': {'enable_dhcp': True}})

                except Exception as e:
                    LOG.error("Failed to enable DHCP on subnet %(subnet)s:"
                              " %(e)s",
                              {'subnet': subnet_id, 'e': e})

        # Add static routes (after all router interfaces and gateways are set)
        for obj_data in objects['router']:
            if 'routes' in obj_data:
                try:
                    plugin.update_router(
                        context, obj_data['id'],
                        {'router': {'routes': obj_data['routes']}})
                except Exception as e:
                    LOG.error("Failed to add routes to router %(rtr)s: "
                              "%(e)s",
                              {'rtr': obj_data['id'], 'e': e})

    LOG.info(">>>Done Creating all objects in NSX-T.")


@admin_utils.output_header
def migrate_v_project_to_t(resource, event, trigger, **kwargs):
    """Migrate 1 project from v to t with all its resources"""

    # filter out the plugins INFO logging
    # TODO(asarfaty): Consider this for all admin utils
    LOG.logger.setLevel(logging.INFO)
    logging.getLogger(None).logger.setLevel(logging.WARN)

    # get the configuration: tenant + public network + from file flag
    usage = ("Usage: nsxadmin -r projects -o %s --property project-id=<> "
             "--property external-net=<NSX-T external network to be used> "
             "<--property from-file=True>" %
             shell.Operations.NSX_MIGRATE_V_V3.value)
    if not kwargs.get('property'):
        LOG.error("Missing parameters: %s", usage)
        return
    properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
    project = properties.get('project-id')
    ext_net_id = properties.get('external-net')
    from_file = properties.get('from-file', 'false').lower() == "true"
    # TODO(asarfaty): get files path
    if not project:
        LOG.error("Missing project-id parameter: %s", usage)
        return
    if not ext_net_id:
        LOG.error("Missing external-net parameter: %s", usage)
        return

    # check if files exist in the current directory
    try:
        filename = get_resource_file_name(project, 'network')
        file = open(filename, 'r')
        if file.read():
            if not from_file:
                from_file = admin_utils.query_yes_no(
                    "Use existing resources files for this project?",
                    default="yes")
        file.close()
    except Exception:
        sys.exc_clear()
        if from_file:
            LOG.error("Cannot run from file: files not found")
            return

    # validate tenant id and public network
    ctx = n_context.get_admin_context()
    mapping = db.get_project_plugin_mapping(ctx.session, project)
    current_plugin = mapping.plugin
    if not mapping:
        LOG.error("Project %s is unknown", project)
        return
    if not from_file and current_plugin != projectpluginmap.NsxPlugins.NSX_V:
        LOG.error("Project %s belongs to plugin %s.", project, mapping.plugin)
        return

    with v3_utils.NsxV3PluginWrapper() as plugin:
        try:
            plugin.get_network(ctx, ext_net_id)
        except exceptions.NetworkNotFound:
            LOG.error("Network %s was not found", ext_net_id)
            return
        if not plugin._network_is_external(ctx, ext_net_id):
            LOG.error("Network %s is not external", ext_net_id)
            return

    if from_file:
        # read resources from files
        objects = read_v_resources_from_files(project)
    else:
        # read all V resources and dump to a file
        objects = read_v_resources_to_files(ctx, project)

    # delete all the V resources (reading it from the files)
    if current_plugin == projectpluginmap.NsxPlugins.NSX_V:
        delete_v_resources(ctx, objects)

    # change the mapping of this tenant to T
    db.update_project_plugin_mapping(ctx.session, project,
                                     projectpluginmap.NsxPlugins.NSX_T)

    # use api replay flag to allow keeping the IDs
    cfg.CONF.set_override('api_replay_mode', True)

    # add resources 1 by one after adapting them to T (api-replay code)
    create_t_resources(ctx, objects, ext_net_id)

    # reset api replay flag to allow keeping the IDs
    cfg.CONF.set_override('api_replay_mode', False)


registry.subscribe(import_projects,
                   constants.PROJECTS,
                   shell.Operations.IMPORT.value)

registry.subscribe(migrate_v_project_to_t,
                   constants.PROJECTS,
                   shell.Operations.NSX_MIGRATE_V_V3.value)



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from oslo_log import log as logging

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib import context as n_context
from neutron_lib import exceptions as nexception
from neutron_lib.plugins import directory
from neutron_vpnaas.services.vpn import service_drivers

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.vpnaas.common_v3 import ipsec_utils

LOG = logging.getLogger(__name__)
IPSEC = 'ipsec'


class RouterWithSNAT(nexception.BadRequest):
    message = _("Router %(router_id)s has a VPN service and cannot enable "
                "SNAT")


class RouterWithOverlapNoSnat(nexception.BadRequest):
    message = _("Router %(router_id)s has a subnet overlapping with a VPN "
                "local subnet, and cannot disable SNAT")


class RouterOverlapping(nexception.BadRequest):
    message = _("Router %(router_id)s interface is overlapping with a VPN "
                "local subnet and cannot be added")


class NSXcommonIPsecVpnDriver(service_drivers.VpnDriver):

    def __init__(self, service_plugin, validator):
        self.vpn_plugin = service_plugin
        self._core_plugin = directory.get_plugin()
        if self._core_plugin.is_tvd_plugin():
            # TVD only supports nsx-T, and not nsx-P
            self._core_plugin = self._core_plugin.get_plugin_by_type(
                projectpluginmap.NsxPlugins.NSX_T)
        super(NSXcommonIPsecVpnDriver, self).__init__(
            service_plugin, validator)

        registry.subscribe(
            self._verify_overlap_subnet, resources.ROUTER_INTERFACE,
            events.BEFORE_CREATE)

    @property
    def l3_plugin(self):
        return self._core_plugin

    @property
    def service_type(self):
        return IPSEC

    def _get_dpd_profile_name(self, connection):
        return (connection['name'] or connection['id'])[:240] + '-dpd-profile'

    def _find_vpn_service_port(self, context, router_id):
        """Look for the neutron port created for the vpnservice of a router"""
        filters = {'device_id': ['router-' + router_id],
                   'device_owner': [ipsec_utils.VPN_PORT_OWNER]}
        ports = self.l3_plugin.get_ports(context, filters=filters)
        if ports:
            return ports[0]

    def _get_service_local_address(self, context, vpnservice):
        """Find/Allocate a port on the external network
        to allocate the ip to be used as the local ip of this service
        """
        router_id = vpnservice['router_id']
        # check if this router already have an IP
        port = self._find_vpn_service_port(context, router_id)
        if not port:
            # create a new port, on the external network of the router
            # Note(asarfaty): using a unique device owner and device id to
            # make sure tis port will be ignored in certain queries
            ext_net = vpnservice['router']['gw_port']['network_id']
            port_data = {
                'port': {
                    'network_id': ext_net,
                    'name': 'VPN local address port',
                    'admin_state_up': True,
                    'device_id': 'router-' + router_id,
                    'device_owner': ipsec_utils.VPN_PORT_OWNER,
                    'fixed_ips': constants.ATTR_NOT_SPECIFIED,
                    'mac_address': constants.ATTR_NOT_SPECIFIED,
                    'port_security_enabled': False,
                    'tenant_id': vpnservice['tenant_id']}}
            port = self.l3_plugin.base_create_port(context, port_data)

        # return the port ip(v4) as the local address
        for fixed_ip in port['fixed_ips']:
            if (len(port['fixed_ips']) == 1 or
                netaddr.IPNetwork(fixed_ip['ip_address']).version == 4):
                return fixed_ip['ip_address']

    def _update_status(self, context, vpn_service_id, ipsec_site_conn_id,
                       status, updated_pending_status=True):
        vpn_status = {'id': vpn_service_id,
                      'updated_pending_status': updated_pending_status,
                      'status': status,
                      'ipsec_site_connections': {}}
        if ipsec_site_conn_id:
            ipsec_site_conn = {
                'status': status,
                'updated_pending_status': updated_pending_status}
            vpn_status['ipsec_site_connections'] = {
                ipsec_site_conn_id: ipsec_site_conn}
        status_list = [vpn_status]
        self.service_plugin.update_status_by_agent(context, status_list)

    def _check_subnets_overlap_with_all_conns(self, context, subnets):
        # find all vpn services with connections
        filters = {'status': [constants.ACTIVE, constants.DOWN]}
        connections = self.vpn_plugin.get_ipsec_site_connections(
            context, filters=filters)
        # Check if any of the connections overlap with the given subnets
        for conn in connections:
            local_cidrs = self.validator._get_local_cidrs(context, conn)
            if netaddr.IPSet(subnets) & netaddr.IPSet(local_cidrs):
                return False

        return True

    def _verify_overlap_subnet(self, resource, event, trigger, **kwargs):
        """Upon router interface creation validation overlapping with vpn"""
        router_db = kwargs.get('router_db')
        port = kwargs.get('port')
        if not port or not router_db:
            LOG.warning("NSX V3 VPNaaS ROUTER_INTERFACE BEFORE_CREATE "
                        "callback didn't get all the relevant information")
            return

        if router_db.enable_snat:
            # checking only no-snat routers
            return

        admin_con = n_context.get_admin_context()
        # Get the (ipv4) subnet of the interface
        subnet_id = None
        for fixed_ip in port['fixed_ips']:
            if netaddr.IPNetwork(fixed_ip['ip_address']).version == 4:
                subnet_id = fixed_ip.get('subnet_id')
                break
        if subnet_id:
            subnet = self._core_plugin.get_subnet(admin_con, subnet_id)
            # find all vpn services with connections
            if not self._check_subnets_overlap_with_all_conns(
                admin_con, [subnet['cidr']]):
                raise RouterOverlapping(router_id=kwargs.get('router_id'))



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def profile(func):
    def wrap(*args, **kwargs):
        f_name = '{}.{}'.format(func.__module__, func.__name__)

        started_at = time.time()
        result = func(*args, **kwargs)
        LOG.debug(">>>>>>>>>>>>> Method %(method)s execution time %(time)f",
                  {'method': f_name, 'time': time.time() - started_at})
        return result

    return wrap



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.services.vpnaas.common_v3 import ipsec_utils
from vmware_nsx.services.vpnaas.common_v3 import ipsec_validator
from vmware_nsxlib.v3 import nsx_constants as consts

LOG = logging.getLogger(__name__)


class IPsecV3Validator(ipsec_validator.IPsecCommonValidator):
    """Validator methods for Vmware NSX-V3 VPN support"""
    def __init__(self, service_plugin):
        super(IPsecV3Validator, self).__init__(service_plugin)

    @property
    def nsxlib(self):
        return self._core_plugin.nsxlib

    def check_backend_version(self):
        if not self.nsxlib.feature_supported(consts.FEATURE_IPSEC_VPN):
            # ipsec vpn is not supported
            LOG.warning("VPNaaS is not supported by the NSX backend (version "
                        "%s)",
                        self.nsxlib.get_version())
            self.backend_support = False
        else:
            self.backend_support = True

    def _validate_backend_version(self):
        if not self.backend_support:
            msg = (_("VPNaaS is not supported by the NSX backend "
                     "(version %s)") % self.nsxlib.get_version())
            raise nsx_exc.NsxVpnValidationError(details=msg)

    @property
    def auth_algorithm_map(self):
        return ipsec_utils.AUTH_ALGORITHM_MAP

    @property
    def pfs_map(self):
        return ipsec_utils.PFS_MAP

    def _validate_t0_ha_mode(self, tier0_uuid):
        # TODO(asarfaty): cache this result
        tier0_router = self.nsxlib.logical_router.get(tier0_uuid)
        if (not tier0_router or
            tier0_router.get('high_availability_mode') != 'ACTIVE_STANDBY'):
            msg = _("The router GW should be connected to a TIER-0 router "
                    "with ACTIVE_STANDBY HA mode")
            raise nsx_exc.NsxVpnValidationError(details=msg)

    def _validate_router(self, context, router_id):
        super(IPsecV3Validator, self)._validate_router(context, router_id)

        # Verify that this is a no-snat router
        router_db = self._core_plugin._get_router(context, router_id)
        if router_db.enable_snat:
            msg = _("VPN is supported only for routers with disabled SNAT")
            raise nsx_exc.NsxVpnValidationError(details=msg)



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.common import v3_common_cert
from vmware_nsx.shell import resources as shell

from neutron_lib.callbacks import registry
from oslo_config import cfg


@admin_utils.output_header
def generate_cert(resource, event, trigger, **kwargs):
    """Generate self signed client certificate and private key
    """
    return v3_common_cert.generate_cert(cfg.CONF.nsx_v3, **kwargs)


@admin_utils.output_header
def delete_cert(resource, event, trigger, **kwargs):
    """Delete client certificate and private key """
    return v3_common_cert.delete_cert(cfg.CONF.nsx_v3, **kwargs)


@admin_utils.output_header
def show_cert(resource, event, trigger, **kwargs):
    """Show client certificate details """
    return v3_common_cert.show_cert(cfg.CONF.nsx_v3, **kwargs)


@admin_utils.output_header
def import_cert(resource, event, trigger, **kwargs):
    """Import client certificate that was generated externally"""
    return v3_common_cert.import_cert(cfg.CONF.nsx_v3, **kwargs)


@admin_utils.output_header
def show_nsx_certs(resource, event, trigger, **kwargs):
    """Show client certificates associated with openstack identity in NSX"""
    return v3_common_cert.show_nsx_certs(cfg.CONF.nsx_v3, **kwargs)


registry.subscribe(generate_cert,
                   constants.CERTIFICATE,
                   shell.Operations.GENERATE.value)

registry.subscribe(show_cert,
                   constants.CERTIFICATE,
                   shell.Operations.SHOW.value)

registry.subscribe(delete_cert,
                   constants.CERTIFICATE,
                   shell.Operations.CLEAN.value)

registry.subscribe(import_cert,
                   constants.CERTIFICATE,
                   shell.Operations.IMPORT.value)

registry.subscribe(show_nsx_certs,
                   constants.CERTIFICATE,
                   shell.Operations.NSX_LIST.value)



# Copyright (c) 2017 VMware, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron.tests import base
from neutron_lbaas.services.loadbalancer import data_models as lb_models
from neutron_lib import context
from neutron_lib import exceptions as n_exc

from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import nsx_models
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsx.services.lbaas.nsx_v3.v2 import lb_driver_v2


LB_VIP = '10.0.0.10'
LB_ROUTER_ID = 'router-x'
ROUTER_ID = 'neutron-router-x'
LB_ID = 'xxx-xxx'
LB_TENANT_ID = 'yyy-yyy'
LB_SERVICE_ID = 'service-1'
LB_BINDING = nsx_models.NsxLbaasLoadbalancer(
    loadbalancer_id=LB_ID,
    lb_service_id=LB_SERVICE_ID,
    lb_router_id=LB_ROUTER_ID,
    vip_address=LB_VIP)
LB_BINDING_NO_RTR = nsx_models.NsxLbaasLoadbalancer(
    loadbalancer_id=LB_ID,
    lb_service_id=LB_SERVICE_ID,
    lb_router_id=lb_utils.NO_ROUTER_ID,
    vip_address=LB_VIP)
LB_NETWORK = {'router:external': False,
              'id': 'xxxxx',
              'name': 'network-1'}
LISTENER_ID = 'listener-x'
HTTP_LISTENER_ID = 'listener-http'
HTTPS_LISTENER_ID = 'listener-https'
APP_PROFILE_ID = 'appp-x'
LB_VS_ID = 'vs-x'
LB_APP_PROFILE = {
    "resource_type": "LbHttpProfile",
    "description": "my http profile",
    "id": APP_PROFILE_ID,
    "display_name": "httpprofile1",
    "ntlm": False,
    "request_header_size": 1024,
    "http_redirect_to_https": False,
    "idle_timeout": 1800,
    "x_forwarded_for": "INSERT",
}
LISTENER_BINDING = nsx_models.NsxLbaasListener(loadbalancer_id=LB_ID,
                                               listener_id=LISTENER_ID,
                                               app_profile_id=APP_PROFILE_ID,
                                               lb_vs_id=LB_VS_ID)
POOL_ID = 'ppp-qqq'
LB_POOL_ID = 'pool-xx'
LB_POOL = {
    "display_name": "httppool1",
    "description": "my http pool",
    "id": LB_POOL_ID,
    "algorithm": "ROUND_ROBIN",
}
POOL_BINDING = nsx_models.NsxLbaasPool(loadbalancer_id=LB_ID,
                                       pool_id=POOL_ID,
                                       lb_pool_id=LB_POOL_ID,
                                       lb_vs_id=LB_VS_ID)
MEMBER_ID = 'mmm-mmm'
MEMBER_ADDRESS = '10.0.0.200'
LB_MEMBER = {'display_name': 'member1_' + MEMBER_ID,
             'weight': 1, 'ip_address': MEMBER_ADDRESS, 'port': 80,
             'backup_member': False}
LB_POOL_WITH_MEMBER = {
    "display_name": "httppool1",
    "description": "my http pool",
    "id": LB_POOL_ID,
    "algorithm": "ROUND_ROBIN",
    "members": [
        {
            "display_name": "http-member1",
            "ip_address": MEMBER_ADDRESS,
            "port": "80",
            "weight": "1",
            "admin_state": "ENABLED"
        }
    ]
}
HM_ID = 'hhh-mmm'
LB_MONITOR_ID = 'mmm-ddd'

HM_BINDING = nsx_models.NsxLbaasMonitor(loadbalancer_id=LB_ID,
                                        pool_id=POOL_ID,
                                        hm_id=HM_ID,
                                        lb_monitor_id=LB_MONITOR_ID,
                                        lb_pool_id=LB_POOL_ID)
L7POLICY_ID = 'l7policy-xxx'
LB_RULE_ID = 'lb-rule-xx'
L7RULE_ID = 'l7rule-111'
L7POLICY_BINDING = nsx_models.NsxLbaasL7Policy(l7policy_id=L7POLICY_ID,
                                               lb_vs_id=LB_VS_ID,
                                               lb_rule_id=LB_RULE_ID)
LB_PP_ID = "ppp-ppp"

FAKE_CERT = {'id': 'cert-xyz'}

SERVICE_STATUSES = {
    "virtual_servers": [{
        "virtual_server_id": LB_VS_ID,
        "status": "UP"
    }],
    "service_id": LB_SERVICE_ID,
    "service_status": "UP",
    "pools": [{
        "members": [{
            "port": "80",
            "ip_address": MEMBER_ADDRESS,
            "status": "DOWN"
        }],
        "pool_id": LB_POOL_ID,
        "status": "DOWN"
    }]
}

VS_STATUSES = {
    "results": [{
        "virtual_server_id": LB_VS_ID,
        "status": "UP"
    }]
}


class BaseTestEdgeLbaasV2(base.BaseTestCase):
    def _tested_entity(self):
        return None

    def setUp(self):
        super(BaseTestEdgeLbaasV2, self).setUp()

        self.context = context.get_admin_context()
        self.edge_driver = lb_driver_v2.EdgeLoadbalancerDriverV2()

        self.lbv2_driver = mock.Mock()
        self.core_plugin = mock.Mock()
        base_mgr.LoadbalancerBaseManager._lbv2_driver = self.lbv2_driver
        base_mgr.LoadbalancerBaseManager._core_plugin = self.core_plugin
        self._patch_lb_plugin(self.lbv2_driver, self._tested_entity)
        self._patch_nsxlib_lb_clients(self.core_plugin)

        self.lb = lb_models.LoadBalancer(LB_ID, LB_TENANT_ID, 'lb1', '',
                                         'some-subnet', 'port-id', LB_VIP)
        self.listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                           'listener1', 'Dummy', None, LB_ID,
                                           'HTTP', protocol_port=80,
                                           loadbalancer=self.lb)
        self.https_listener = lb_models.Listener(
            HTTP_LISTENER_ID, LB_TENANT_ID, 'listener2', '', None, LB_ID,
            'HTTPS', protocol_port=443, loadbalancer=self.lb)
        self.terminated_https_listener = lb_models.Listener(
            HTTPS_LISTENER_ID, LB_TENANT_ID, 'listener3', '', None, LB_ID,
            'TERMINATED_HTTPS', protocol_port=443, loadbalancer=self.lb)
        self.pool = lb_models.Pool(POOL_ID, LB_TENANT_ID, 'pool1', '',
                                   None, 'HTTP', 'ROUND_ROBIN',
                                   loadbalancer_id=LB_ID,
                                   listener=self.listener,
                                   listeners=[self.listener],
                                   loadbalancer=self.lb)
        self.sess_persistence = lb_models.SessionPersistence(
            POOL_ID, 'HTTP_COOKIE', 'meh_cookie')
        self.pool_persistency = lb_models.Pool(POOL_ID, LB_TENANT_ID,
                                   'pool1', '', None, 'HTTP',
                                   'ROUND_ROBIN', loadbalancer_id=LB_ID,
                                   listener=self.listener,
                                   listeners=[self.listener],
                                   loadbalancer=self.lb,
                                   session_persistence=self.sess_persistence)
        self.member = lb_models.Member(MEMBER_ID, LB_TENANT_ID, POOL_ID,
                                       MEMBER_ADDRESS, 80, 1, pool=self.pool,
                                       name='member1')
        self.hm = lb_models.HealthMonitor(HM_ID, LB_TENANT_ID, 'PING', 3, 3,
                                          1, pool=self.pool, name='hm1')
        self.hm_http = lb_models.HealthMonitor(HM_ID, LB_TENANT_ID, 'HTTP',
                                               3, 3, 1, pool=self.pool,
                                               http_method='GET',
                                               url_path="/meh", name='hm2')

        self.l7policy = lb_models.L7Policy(L7POLICY_ID, LB_TENANT_ID,
                                           name='policy-test',
                                           description='policy-desc',
                                           listener_id=LISTENER_ID,
                                           action='REDIRECT_TO_POOL',
                                           redirect_pool_id=POOL_ID,
                                           listener=self.listener,
                                           position=1)
        self.l7rule = lb_models.L7Rule(L7RULE_ID, LB_TENANT_ID,
                                       l7policy_id=L7POLICY_ID,
                                       compare_type='EQUAL_TO',
                                       invert=False,
                                       type='HEADER',
                                       key='key1',
                                       value='val1',
                                       policy=self.l7policy)

    def tearDown(self):
        self._unpatch_lb_plugin(self.lbv2_driver, self._tested_entity)
        super(BaseTestEdgeLbaasV2, self).tearDown()

    def _patch_lb_plugin(self, lb_plugin, manager):
        self.real_manager = getattr(lb_plugin, manager)
        lb_manager = mock.patch.object(lb_plugin, manager).start()
        mock.patch.object(lb_manager, 'create').start()
        mock.patch.object(lb_manager, 'update').start()
        mock.patch.object(lb_manager, 'delete').start()
        mock.patch.object(lb_manager, 'successful_completion').start()

    def _patch_nsxlib_lb_clients(self, core_plugin):
        nsxlib = mock.patch.object(core_plugin, 'nsxlib').start()
        load_balancer = mock.patch.object(nsxlib, 'load_balancer').start()
        self.service_client = mock.patch.object(load_balancer,
                                                'service').start()
        self.app_client = mock.patch.object(load_balancer,
                                            'application_profile').start()
        self.vs_client = mock.patch.object(load_balancer,
                                           'virtual_server').start()
        self.pool_client = mock.patch.object(load_balancer,
                                             'pool').start()
        self.monitor_client = mock.patch.object(load_balancer,
                                                'monitor').start()
        self.rule_client = mock.patch.object(load_balancer,
                                             'rule').start()
        self.pp_client = mock.patch.object(load_balancer,
                                           'persistence_profile').start()
        self.tm_client = mock.patch.object(nsxlib,
                                           'trust_management').start()
        self.nsxlib = nsxlib

    def _unpatch_lb_plugin(self, lb_plugin, manager):
        setattr(lb_plugin, manager, self.real_manager)


class TestEdgeLbaasV2Loadbalancer(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2Loadbalancer, self).setUp()

    @property
    def _tested_entity(self):
        return 'load_balancer'

    def test_create(self):
        neutron_router = {'id': ROUTER_ID, 'name': 'dummy',
                          'external_gateway_info': {'external_fixed_ips': []}}
        with mock.patch.object(lb_utils, 'validate_lb_subnet'
                               ) as mock_validate_lb_subnet,\
            mock.patch.object(lb_utils, 'get_router_from_network',
                              return_value=ROUTER_ID),\
            mock.patch.object(self.core_plugin, 'get_router',
                              return_value=neutron_router), \
            mock.patch.object(self.core_plugin, '_find_router_gw_subnets',
                              return_value=[]),\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'),\
            mock.patch.object(nsx_db, 'get_nsx_router_id',
                              return_value=LB_ROUTER_ID),\
            mock.patch.object(self.service_client, 'get_router_lb_service',
                              return_value=None),\
            mock.patch.object(self.service_client, 'create',
                              return_value={'id': LB_SERVICE_ID}
                              ) as create_service,\
            mock.patch.object(nsx_db, 'add_nsx_lbaas_loadbalancer_binding'
                              ) as add_binding:
            mock_validate_lb_subnet.return_value = True

            self.edge_driver.loadbalancer.create(self.context, self.lb)

            mock_successful_completion = (
                self.lbv2_driver.load_balancer.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.lb,
                                                          delete=False)
            add_binding.assert_called_once_with(mock.ANY, LB_ID, LB_SERVICE_ID,
                                                LB_ROUTER_ID, LB_VIP)
            create_service.assert_called_once()

    def test_create_service_exists(self):
        with mock.patch.object(lb_utils, 'validate_lb_subnet'
                               ) as mock_validate_lb_subnet,\
            mock.patch.object(lb_utils, 'get_router_from_network'),\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'),\
            mock.patch.object(nsx_db, 'get_nsx_router_id',
                              return_value=LB_ROUTER_ID),\
            mock.patch.object(self.service_client, 'get_router_lb_service',
                              return_value={'id': LB_SERVICE_ID}),\
            mock.patch.object(self.service_client,
                              'create') as create_service,\
            mock.patch.object(nsx_db, 'add_nsx_lbaas_loadbalancer_binding'
                              ) as add_binding:
            mock_validate_lb_subnet.return_value = True

            self.edge_driver.loadbalancer.create(self.context, self.lb)

            mock_successful_completion = (
                self.lbv2_driver.load_balancer.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.lb,
                                                          delete=False)
            add_binding.assert_called_once_with(mock.ANY, LB_ID, LB_SERVICE_ID,
                                                LB_ROUTER_ID, LB_VIP)
            create_service.assert_not_called()

    def test_create_external_vip(self):
        with mock.patch.object(lb_utils, 'validate_lb_subnet'
                               ) as mock_validate_lb_subnet,\
            mock.patch.object(lb_utils, 'get_router_from_network',
                              return_value=None),\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'),\
            mock.patch.object(nsx_db, 'get_nsx_router_id'),\
            mock.patch.object(self.service_client, 'get_router_lb_service',
                              return_value=None),\
            mock.patch.object(self.service_client, 'create',
                              return_value={'id': LB_SERVICE_ID}
                              ) as create_service,\
            mock.patch.object(nsx_db, 'add_nsx_lbaas_loadbalancer_binding'
                              ) as add_binding:
            mock_validate_lb_subnet.return_value = True

            self.edge_driver.loadbalancer.create(self.context, self.lb)

            mock_successful_completion = (
                self.lbv2_driver.load_balancer.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.lb,
                                                          delete=False)
            add_binding.assert_called_once_with(mock.ANY, LB_ID, LB_SERVICE_ID,
                                                lb_utils.NO_ROUTER_ID, LB_VIP)
            create_service.assert_called_once()

    def test_update(self):
        new_lb = lb_models.LoadBalancer(LB_ID, 'yyy-yyy', 'lb1-new',
                                        'new-description', 'some-subnet',
                                        'port-id', LB_VIP)

        self.edge_driver.loadbalancer.update(self.context, self.lb, new_lb)

        mock_successful_completion = (
            self.lbv2_driver.load_balancer.successful_completion)
        mock_successful_completion.assert_called_with(self.context, new_lb,
                                                      delete=False)

    def test_delete(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                               ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'get'
                              ) as mock_get_lb_service, \
            mock.patch.object(self.service_client, 'delete'
                              ) as mock_delete_lb_service, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(nsx_db, 'delete_nsx_lbaas_loadbalancer_binding'
                              ) as mock_delete_lb_binding:
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_lb_service.return_value = {'id': LB_SERVICE_ID}

            self.edge_driver.loadbalancer.delete(self.context, self.lb)

            mock_delete_lb_service.assert_called_with(LB_SERVICE_ID)
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID
            mock_delete_lb_binding.assert_called_with(
                self.context.session, LB_ID)
            mock_successful_completion = (
                self.lbv2_driver.load_balancer.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.lb,
                                                          delete=True)

    def test_stats(self):
        pass

    def test_refresh(self):
        pass

    def test_status_update(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                               ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'get_status'
                              ) as mock_get_lb_service_status, \
            mock.patch.object(self.service_client, 'get_virtual_servers_status'
                              ) as mock_get_vs_status, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding_by_lb_pool'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.pool_client, 'get'
                              ) as mock_get_pool, \
            mock.patch.object(nsx_db,
                              'get_nsx_lbaas_listener_binding_by_lb_and_vs'
                              ) as mock_get_listener_binding:
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_listener_binding.return_value = LISTENER_BINDING
            mock_get_lb_service_status.return_value = SERVICE_STATUSES
            mock_get_vs_status.return_value = VS_STATUSES
            mock_get_pool.return_value = LB_POOL_WITH_MEMBER
            statuses = self.edge_driver.loadbalancer.get_operating_status(
                self.context, self.lb.id, with_members=True)
            self.assertEqual(1, len(statuses['loadbalancers']))
            self.assertEqual('ONLINE', statuses['loadbalancers'][0]['status'])
            self.assertEqual(1, len(statuses['pools']))
            self.assertEqual('OFFLINE', statuses['pools'][0]['status'])
            self.assertEqual(1, len(statuses['listeners']))
            self.assertEqual('ONLINE', statuses['listeners'][0]['status'])
            self.assertEqual(1, len(statuses['members']))
            self.assertEqual('OFFLINE', statuses['members'][0]['status'])


class TestEdgeLbaasV2Listener(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2Listener, self).setUp()

    @property
    def _tested_entity(self):
        return 'listener'

    def _create_listener(self, protocol='HTTP'):
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(self.app_client, 'create'
                              ) as mock_create_app_profile, \
            mock.patch.object(self.vs_client, 'create'
                              ) as mock_create_virtual_server, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'add_virtual_server'
                              ) as mock_add_virtual_server, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_listener_binding'
                              ) as mock_add_listener_binding:
            mock_get_floatingips.return_value = []
            mock_create_app_profile.return_value = {'id': APP_PROFILE_ID}
            mock_create_virtual_server.return_value = {'id': LB_VS_ID}
            mock_get_lb_binding.return_value = LB_BINDING
            listener = self.listener
            if protocol == 'HTTPS':
                listener = self.https_listener

            self.edge_driver.listener.create(self.context, listener)

            mock_add_virtual_server.assert_called_with(LB_SERVICE_ID,
                                                       LB_VS_ID)
            mock_add_listener_binding.assert_called_with(
                self.context.session, LB_ID, listener.id, APP_PROFILE_ID,
                LB_VS_ID)
            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          listener,
                                                          delete=False)

    def test_create_http_listener(self):
        self._create_listener()

    def test_create_https_listener(self):
        self._create_listener(protocol='HTTPS')

    def test_create_terminated_https(self):
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(self.tm_client, 'create_cert'
                              ) as mock_create_cert, \
            mock.patch.object(self.app_client, 'create'
                              ) as mock_create_app_profile, \
            mock.patch.object(self.vs_client, 'create'
                              ) as mock_create_virtual_server, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'add_virtual_server'
                              ) as mock_add_virtual_server, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_listener_binding'
                              ) as mock_add_listener_binding:
            mock_get_floatingips.return_value = []
            mock_create_cert.return_value = FAKE_CERT['id']
            mock_create_app_profile.return_value = {'id': APP_PROFILE_ID}
            mock_create_virtual_server.return_value = {'id': LB_VS_ID}
            mock_get_lb_binding.return_value = LB_BINDING

            self.edge_driver.listener.create(self.context,
                                             self.terminated_https_listener)
            mock_add_virtual_server.assert_called_with(LB_SERVICE_ID,
                                                       LB_VS_ID)
            mock_add_listener_binding.assert_called_with(
                self.context.session, LB_ID, HTTPS_LISTENER_ID, APP_PROFILE_ID,
                LB_VS_ID)

            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(
                self.context, self.terminated_https_listener,
                delete=False)

    def test_create_listener_with_default_pool(self):
        listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                      'listener1', 'Dummy', self.pool.id,
                                      LB_ID, 'HTTP', protocol_port=80,
                                      loadbalancer=self.lb,
                                      default_pool=self.pool)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(self.app_client, 'create'
                              ) as mock_create_app_profile, \
            mock.patch.object(self.vs_client, 'create'
                              ) as mock_create_virtual_server, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'add_virtual_server'
                              ) as mock_add_virtual_server, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_listener_binding'
                              ) as mock_add_listener_binding,\
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'),\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding:
            mock_get_floatingips.return_value = []
            mock_create_app_profile.return_value = {'id': APP_PROFILE_ID}
            mock_create_virtual_server.return_value = {'id': LB_VS_ID}
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_pool_binding.return_value = None

            self.edge_driver.listener.create(self.context, listener)

            mock_add_virtual_server.assert_called_with(LB_SERVICE_ID,
                                                       LB_VS_ID)
            mock_add_listener_binding.assert_called_with(
                self.context.session, LB_ID, LISTENER_ID, APP_PROFILE_ID,
                LB_VS_ID)
            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          listener,
                                                          delete=False)

    def test_create_listener_with_used_default_pool(self):
        listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                      'listener1', 'Dummy', self.pool.id,
                                      LB_ID, 'HTTP', protocol_port=80,
                                      loadbalancer=self.lb,
                                      default_pool=self.pool)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding:
            mock_get_floatingips.return_value = []
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.assertRaises(n_exc.BadRequest,
                              self.edge_driver.listener.create,
                              self.context, listener)

    def test_create_listener_with_session_persistence(self):
        listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                      'listener1', 'Dummy',
                                      self.pool_persistency.id,
                                      LB_ID, 'HTTP', protocol_port=80,
                                      loadbalancer=self.lb,
                                      default_pool=self.pool_persistency)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(self.app_client, 'create'
                              ) as mock_create_app_profile, \
            mock.patch.object(self.vs_client, 'create'
                              ) as mock_create_virtual_server, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'add_virtual_server'
                              ) as mock_add_virtual_server, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_listener_binding'
                              ) as mock_add_listener_binding,\
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'),\
            mock.patch.object(self.pp_client, 'create'
                              ) as mock_create_pp, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding:
            mock_get_floatingips.return_value = []
            mock_create_app_profile.return_value = {'id': APP_PROFILE_ID}
            mock_create_virtual_server.return_value = {'id': LB_VS_ID}
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_pool_binding.return_value = None

            self.edge_driver.listener.create(self.context, listener)

            mock_add_virtual_server.assert_called_with(LB_SERVICE_ID,
                                                       LB_VS_ID)
            mock_add_listener_binding.assert_called_with(
                self.context.session, LB_ID, LISTENER_ID, APP_PROFILE_ID,
                LB_VS_ID)
            mock_create_pp.assert_called_once()
            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          listener,
                                                          delete=False)

    def test_create_listener_with_session_persistence_fail(self):
        listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                      'listener1', 'Dummy',
                                      self.pool_persistency.id,
                                      LB_ID, 'TCP', protocol_port=80,
                                      loadbalancer=self.lb,
                                      default_pool=self.pool_persistency)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding:
            mock_get_floatingips.return_value = []
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_pool_binding.return_value = None

            self.assertRaises(n_exc.BadRequest,
                              self.edge_driver.listener.create,
                              self.context, listener)

    def test_update(self):
        new_listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                          'listener1-new', 'new-description',
                                          None, LB_ID, protocol_port=80,
                                          loadbalancer=self.lb)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                              ) as mock_get_listener_binding:
            mock_get_floatingips.return_value = []
            mock_get_listener_binding.return_value = LISTENER_BINDING

            self.edge_driver.listener.update(self.context, self.listener,
                                             new_listener)

            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_listener,
                                                          delete=False)

    def test_update_with_default_pool(self):
        new_listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                          'listener1-new', 'new-description',
                                          self.pool, LB_ID, protocol_port=80,
                                          loadbalancer=self.lb,
                                          default_pool=self.pool)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                              ) as mock_get_listener_binding,\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding,\
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'):
            mock_get_floatingips.return_value = []
            mock_get_listener_binding.return_value = LISTENER_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.listener.update(self.context, self.listener,
                                             new_listener)

            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_listener,
                                                          delete=False)

    def test_update_with_session_persistence(self):
        new_listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                          'listener1-new', 'new-description',
                                          self.pool_persistency.id,
                                          LB_ID, protocol='HTTP',
                                          protocol_port=80,
                                          loadbalancer=self.lb,
                                          default_pool=self.pool_persistency)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                              ) as mock_get_listener_binding,\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding,\
            mock.patch.object(self.vs_client, 'update',
                              return_value={'id': LB_VS_ID}), \
            mock.patch.object(self.pp_client, 'create'
                              ) as mock_create_pp, \
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'):
            mock_get_floatingips.return_value = []
            mock_get_listener_binding.return_value = LISTENER_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.listener.update(self.context, self.listener,
                                             new_listener)
            mock_create_pp.assert_called_once()
            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_listener,
                                                          delete=False)

    def test_update_with_session_persistence_change(self):
        old_listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                          'listener1', 'description',
                                          self.pool_persistency.id,
                                          LB_ID, protocol='HTTP',
                                          protocol_port=80,
                                          loadbalancer=self.lb,
                                          default_pool=self.pool_persistency)
        sess_persistence = lb_models.SessionPersistence(
            POOL_ID, 'SOURCE_IP')
        pool_persistency = lb_models.Pool('new_pool_id', LB_TENANT_ID,
                                   'pool1', '', None, 'HTTP',
                                   'ROUND_ROBIN', loadbalancer_id=LB_ID,
                                   listener=self.listener,
                                   listeners=[self.listener],
                                   loadbalancer=self.lb,
                                   session_persistence=sess_persistence)
        new_listener = lb_models.Listener(LISTENER_ID, LB_TENANT_ID,
                                          'listener1-new', 'new-description',
                                          pool_persistency.id,
                                          LB_ID, protocol='HTTP',
                                          protocol_port=80,
                                          loadbalancer=self.lb,
                                          default_pool=pool_persistency)
        with mock.patch.object(self.core_plugin, 'get_floatingips'
                               ) as mock_get_floatingips, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                              ) as mock_get_listener_binding,\
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.pp_client, 'create'
                              ) as mock_create_pp, \
            mock.patch.object(self.pp_client, 'delete'
                              ) as mock_delete_pp, \
            mock.patch.object(lb_utils, 'get_pool_tags'
                              ) as mock_get_pool_tags, \
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'):
            mock_get_pool_tags.return_value = []
            mock_get_floatingips.return_value = []
            mock_get_listener_binding.return_value = LISTENER_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.listener.update(
                self.context, old_listener,
                new_listener)
            mock_create_pp.assert_called_once_with(
                display_name='persistence_pool1_new_p...ol_id',
                resource_type='LbSourceIpPersistenceProfile',
                tags=mock.ANY)
            # No reason to check parameters here, it's
            # all mocked out
            mock_delete_pp.assert_called_once()

    def test_delete(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                               ) as mock_get_listener_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(self.service_client, 'get'
                              ) as mock_get_lb_service, \
            mock.patch.object(self.service_client, 'remove_virtual_server'
                              ) as mock_remove_virtual_server, \
            mock.patch.object(self.app_client, 'delete'
                              ) as mock_delete_app_profile, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(self.vs_client, 'delete'
                              ) as mock_delete_virtual_server, \
            mock.patch.object(nsx_db, 'delete_nsx_lbaas_listener_binding',
                              ) as mock_delete_listener_binding:
            mock_get_listener_binding.return_value = LISTENER_BINDING
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_lb_service.return_value = {
                'id': LB_SERVICE_ID,
                'virtual_server_ids': [LB_VS_ID]}

            self.edge_driver.listener.delete(self.context, self.listener)

            mock_remove_virtual_server.assert_called_with(LB_SERVICE_ID,
                                                          LB_VS_ID)
            mock_delete_virtual_server.assert_called_with(LB_VS_ID)
            mock_delete_app_profile.assert_called_with(APP_PROFILE_ID)

            mock_delete_listener_binding.assert_called_with(
                self.context.session, LB_ID, LISTENER_ID)

            mock_successful_completion = (
                self.lbv2_driver.listener.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.listener,
                                                          delete=True)


class TestEdgeLbaasV2Pool(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2Pool, self).setUp()

    @property
    def _tested_entity(self):
        return 'pool'

    def test_create(self):
        with mock.patch.object(self.pool_client, 'create'
                               ) as mock_create_pool, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_pool_binding'
                              ) as mock_add_pool_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                              ) as mock_get_listener_binding, \
            mock.patch.object(self.pp_client, 'create'
                              ) as mock_create_pp, \
            mock.patch.object(self.vs_client, 'update', return_value=None
                              ) as mock_vs_update, \
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'
                              ) as mock_update_pool_binding:
            mock_create_pool.return_value = {'id': LB_POOL_ID}
            mock_get_listener_binding.return_value = LISTENER_BINDING

            self.edge_driver.pool.create(self.context, self.pool)

            mock_add_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, LB_POOL_ID)
            mock_create_pp.assert_not_called()
            mock_vs_update.assert_called_once_with(
                LB_VS_ID, pool_id=LB_POOL_ID, persistence_profile_id=None)
            mock_update_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, LB_VS_ID)
            mock_successful_completion = (
                self.lbv2_driver.pool.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.pool,
                                                          delete=False)

    def _test_create_with_persistency(self, vs_data, verify_func):
        with mock.patch.object(self.pool_client, 'create'
                               ) as mock_create_pool, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_pool_binding'
                              ) as mock_add_pool_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                              ) as mock_get_listener_binding, \
            mock.patch.object(self.pp_client, 'create'
                              ) as mock_create_pp, \
            mock.patch.object(self.pp_client, 'update', return_value=None,
                              ) as mock_update_pp, \
            mock.patch.object(self.vs_client, 'get'
                              ) as mock_vs_get, \
            mock.patch.object(self.vs_client, 'update', return_value=None
                              ) as mock_vs_update, \
            mock.patch.object(nsx_db, 'update_nsx_lbaas_pool_binding'
                              ) as mock_update_pool_binding:

            mock_vs_get.return_value = vs_data
            mock_create_pool.return_value = {'id': LB_POOL_ID}
            mock_create_pp.return_value = {'id': LB_PP_ID}
            mock_get_listener_binding.return_value = LISTENER_BINDING

            self.edge_driver.pool.create(self.context, self.pool_persistency)

            mock_add_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, LB_POOL_ID)
            verify_func(mock_create_pp, mock_update_pp,
                        mock_update_pool_binding, mock_vs_update)
            mock_successful_completion = (
                self.lbv2_driver.pool.successful_completion)
            mock_successful_completion.assert_called_with(
                self.context, self.pool_persistency, delete=False)

    def test_create_with_persistency(self):

        def verify_func(mock_create_pp, mock_update_pp,
                        mock_update_pool_binding, mock_vs_update):
            mock_create_pp.assert_called_once_with(
                resource_type='LbCookiePersistenceProfile',
                cookie_mode='INSERT',
                cookie_name='meh_cookie',
                display_name=mock.ANY,
                tags=mock.ANY)
            mock_update_pp.assert_not_called()
            mock_update_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, LB_VS_ID)
            mock_vs_update.assert_called_once_with(
                LB_VS_ID, pool_id=LB_POOL_ID, persistence_profile_id=LB_PP_ID)

        vs_data = {'id': LB_VS_ID}
        self._test_create_with_persistency(vs_data, verify_func)

    def test_create_with_persistency_existing_profile(self):
        def verify_func(mock_create_pp, mock_update_pp,
                        mock_update_pool_binding, mock_vs_update):
            mock_create_pp.assert_not_called()
            mock_update_pp.assert_called_once_with(
                LB_PP_ID,
                resource_type='LbCookiePersistenceProfile',
                cookie_mode='INSERT',
                cookie_name='meh_cookie',
                display_name=mock.ANY,
                tags=mock.ANY)
            mock_update_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, LB_VS_ID)
            mock_vs_update.assert_called_once_with(
                LB_VS_ID, pool_id=LB_POOL_ID, persistence_profile_id=LB_PP_ID)

        vs_data = {'id': LB_VS_ID,
                   'persistence_profile_id': LB_PP_ID}
        self._test_create_with_persistency(vs_data, verify_func)

    def test_create_with_persistency_no_listener(self):
        def verify_func(mock_create_pp, mock_update_pp,
                        mock_update_pool_binding, mock_vs_update):
            mock_create_pp.assert_not_called()
            mock_update_pp.assert_not_called()
            mock_update_pool_binding.assert_not_called()
            mock_vs_update.assert_not_called()

        vs_data = {'id': LB_VS_ID,
                   'persistence_profile_id': LB_PP_ID}
        self.pool_persistency.listener = None
        self.pool_persistency.listeners = []
        self._test_create_with_persistency(vs_data, verify_func)

    def test_create_multiple_listeners(self):
        """Verify creation will fail if multiple listeners are set"""
        pool = lb_models.Pool(POOL_ID, LB_TENANT_ID, 'pool1', '',
                              None, 'HTTP', 'ROUND_ROBIN',
                              loadbalancer_id=LB_ID,
                              listeners=[self.listener,
                                         self.https_listener],
                              loadbalancer=self.lb)
        self.assertRaises(n_exc.BadRequest,
                          self.edge_driver.pool.create,
                          self.context, pool)

    def test_update(self):
        new_pool = lb_models.Pool(POOL_ID, LB_TENANT_ID, 'pool-name', '',
                                  None, 'HTTP', 'LEAST_CONNECTIONS',
                                  listener=self.listener)
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding:
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.pool.update(self.context, self.pool, new_pool)

            mock_successful_completion = (
                self.lbv2_driver.pool.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_pool,
                                                          delete=False)

    def test_update_multiple_listeners(self):
        """Verify update action will fail if multiple listeners are set"""
        new_pool = lb_models.Pool(POOL_ID, LB_TENANT_ID, 'pool1', '',
                                  None, 'HTTP', 'ROUND_ROBIN',
                                  loadbalancer_id=LB_ID,
                                  listeners=[self.listener,
                                             self.https_listener],
                                  loadbalancer=self.lb)
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding:
            mock_get_pool_binding.return_value = POOL_BINDING
            self.assertRaises(n_exc.BadRequest,
                              self.edge_driver.pool.update,
                              self.context, self.pool, new_pool)

    def _test_update_with_persistency(self, vs_data, old_pool, new_pool,
                                      verify_func):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding, \
            mock.patch.object(self.pp_client, 'create'
                              ) as mock_create_pp, \
            mock.patch.object(self.pp_client, 'update', return_value=None,
                              ) as mock_update_pp, \
            mock.patch.object(self.pp_client, 'delete', return_value=None,
                              ) as mock_delete_pp, \
            mock.patch.object(self.vs_client, 'get'
                              ) as mock_vs_get, \
            mock.patch.object(self.vs_client, 'update', return_value=None
                              ) as mock_vs_update:

            mock_vs_get.return_value = vs_data
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_create_pp.return_value = {'id': LB_PP_ID}

            self.edge_driver.pool.update(self.context, old_pool, new_pool)

            verify_func(mock_create_pp, mock_update_pp,
                        mock_delete_pp, mock_vs_update)
            mock_successful_completion = (
                self.lbv2_driver.pool.successful_completion)
            mock_successful_completion.assert_called_with(
                self.context, new_pool, delete=False)

    def test_update_with_persistency(self):

        def verify_func(mock_create_pp, mock_update_pp,
                        mock_delete_pp, mock_vs_update):
            mock_create_pp.assert_called_once_with(
                resource_type='LbCookiePersistenceProfile',
                cookie_mode='INSERT',
                cookie_name='meh_cookie',
                display_name=mock.ANY,
                tags=mock.ANY)
            mock_update_pp.assert_not_called()
            mock_delete_pp.assert_not_called()
            mock_vs_update.assert_called_once_with(
                LB_VS_ID, pool_id=LB_POOL_ID, persistence_profile_id=LB_PP_ID)

        vs_data = {'id': LB_VS_ID}
        self._test_update_with_persistency(vs_data, self.pool,
                                           self.pool_persistency, verify_func)

    def test_update_remove_persistency(self):
        def verify_func(mock_create_pp, mock_update_pp,
                        mock_delete_pp, mock_vs_update):
            mock_create_pp.assert_not_called()
            mock_update_pp.assert_not_called()
            mock_delete_pp.assert_called_with(LB_PP_ID)
            mock_vs_update.assert_called_once_with(
                LB_VS_ID, pool_id=LB_POOL_ID, persistence_profile_id=None)

        vs_data = {'id': LB_VS_ID,
                   'persistence_profile_id': LB_PP_ID}
        self._test_update_with_persistency(vs_data, self.pool_persistency,
                                           self.pool, verify_func)

    def test_delete(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding, \
            mock.patch.object(self.vs_client, 'update', return_value=None
                              ) as mock_update_virtual_server, \
            mock.patch.object(self.pool_client, 'delete'
                              ) as mock_delete_pool, \
            mock.patch.object(nsx_db, 'delete_nsx_lbaas_pool_binding'
                              ) as mock_delete_pool_binding, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding:
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID
            mock_get_lb_binding.return_value = None

            self.edge_driver.pool.delete(self.context, self.pool)

            mock_update_virtual_server.assert_called_with(
                LB_VS_ID, persistence_profile_id=None, pool_id=None)
            mock_delete_pool.assert_called_with(LB_POOL_ID)
            mock_delete_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID)

            mock_successful_completion = (
                self.lbv2_driver.pool.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.pool,
                                                          delete=True)

    def test_delete_with_persistency(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding, \
            mock.patch.object(self.vs_client, 'get'
                              ) as mock_vs_get, \
            mock.patch.object(self.vs_client, 'update', return_value=None
                              ) as mock_update_virtual_server, \
            mock.patch.object(self.pool_client, 'delete'
                              ) as mock_delete_pool, \
            mock.patch.object(self.pp_client, 'delete', return_value=None,
                              ) as mock_delete_pp, \
            mock.patch.object(nsx_db, 'delete_nsx_lbaas_pool_binding'
                              ) as mock_delete_pool_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding:
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_lb_binding.return_value = None
            mock_vs_get.return_value = {'id': LB_VS_ID,
                                        'persistence_profile_id': LB_PP_ID}

            self.edge_driver.pool.delete(self.context, self.pool_persistency)

            mock_delete_pp.assert_called_once_with(LB_PP_ID)
            mock_update_virtual_server.assert_called_once_with(
                LB_VS_ID, persistence_profile_id=None, pool_id=None)
            mock_delete_pool.assert_called_with(LB_POOL_ID)
            mock_delete_pool_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID)

            mock_successful_completion = (
                self.lbv2_driver.pool.successful_completion)
            mock_successful_completion.assert_called_with(
                self.context, self.pool_persistency, delete=True)

    def _verify_create(self, res_type, cookie_name, cookie_mode,
                       mock_create_pp, mock_update_pp):
        if cookie_name:
            mock_create_pp.assert_called_once_with(
                resource_type=res_type,
                cookie_name=cookie_name,
                cookie_mode=cookie_mode,
                display_name=mock.ANY,
                tags=mock.ANY)
        else:
            mock_create_pp.assert_called_once_with(
                resource_type=res_type,
                display_name=mock.ANY,
                tags=mock.ANY)
        # Compare tags - kw args are the last item of a mock call tuple
        self.assertItemsEqual(mock_create_pp.mock_calls[0][-1]['tags'],
            [{'scope': 'os-lbaas-lb-id', 'tag': 'xxx-xxx'},
                {'scope': 'os-lbaas-lb-name', 'tag': 'lb1'},
                {'scope': 'os-lbaas-listener-id', 'tag': 'listener-x'}])
        mock_update_pp.assert_not_called()

    def _verify_update(self, res_type, cookie_name, cookie_mode,
                       mock_create_pp, mock_update_pp):
        if cookie_name:
            mock_update_pp.assert_called_once_with(
                LB_PP_ID,
                resource_type=res_type,
                cookie_name=cookie_name,
                cookie_mode=cookie_mode,
                display_name=mock.ANY,
                tags=mock.ANY)
        else:
            mock_update_pp.assert_called_once_with(
                LB_PP_ID,
                resource_type=res_type,
                display_name=mock.ANY,
                tags=mock.ANY)
        # Compare tags - kw args are the last item of a mock call tuple
        self.assertItemsEqual(mock_update_pp.mock_calls[0][-1]['tags'],
            [{'scope': 'os-lbaas-lb-id', 'tag': 'xxx-xxx'},
             {'scope': 'os-lbaas-lb-name', 'tag': 'lb1'},
             {'scope': 'os-lbaas-listener-id', 'tag': 'listener-x'}])
        mock_create_pp.assert_not_called()

    def _verify_switch(self, res_type, cookie_name, cookie_mode,
                       mock_create_pp, mock_update_pp):
        if cookie_name:
            mock_create_pp.assert_called_once_with(
                resource_type=res_type,
                cookie_name=cookie_name,
                cookie_mode=cookie_mode,
                display_name=mock.ANY,
                tags=mock.ANY)
        else:
            mock_create_pp.assert_called_once_with(
                LB_PP_ID,
                resource_type=res_type,
                display_name=mock.ANY,
                tags=mock.ANY)
        # Compare tags - kw args are the last item of a mock call tuple
        self.assertItemsEqual(mock_create_pp.mock_calls[0][-1]['tags'],
            [{'scope': 'os-lbaas-lb-id', 'tag': 'xxx-xxx'},
             {'scope': 'os-lbaas-lb-name', 'tag': 'lb1'},
             {'scope': 'os-lbaas-listener-id', 'tag': 'listener-x'}])

    def _verify_delete(self, res_type, cookie_name, cookie_mode,
                       mock_create_pp, mock_update_pp):
        # do not check delete mock as deletion is not done in
        # setup_session_persistence
        mock_create_pp.assert_not_called()
        mock_update_pp.assert_not_called()

    def _test_setup_session_persistence(self, session_persistence,
                                        res_type, vs_data, verify_func,
                                        cookie_name=None, cookie_mode=None,
                                        switch_type=False):
        with mock.patch.object(self.pp_client, 'create'
                               ) as mock_create_pp, \
            mock.patch.object(self.pp_client, 'update', return_value=None,
                              ) as mock_update_pp:

            mock_create_pp.return_value = {'id': LB_PP_ID}
            self.pool.session_persistence = session_persistence
            pool_dict = self.edge_driver.pool.translator(self.pool)
            list_dict = self.edge_driver.listener.translator(self.listener)
            pp_id, post_func = lb_utils.setup_session_persistence(
                self.nsxlib, pool_dict, [], switch_type, list_dict,
                vs_data)

            if session_persistence:
                self.assertEqual(LB_PP_ID, pp_id)
            else:
                self.assertIsNone(pp_id)

            if not session_persistence or switch_type:
                # Also verify post_func for delete
                self.assertEqual((self.nsxlib, LB_PP_ID,),
                                 post_func.args)
            verify_func(res_type, cookie_name, cookie_mode,
                        mock_create_pp, mock_update_pp)

    def test_setup_session_persistence_sourceip_new_profile(self):
        sess_persistence = lb_models.SessionPersistence(POOL_ID, 'SOURCE_IP')
        res_type = 'LbSourceIpPersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type, {'id': LB_VS_ID}, self._verify_create)

    def test_setup_session_persistence_httpcookie_new_profile(self):
        sess_persistence = lb_models.SessionPersistence(
            POOL_ID, 'HTTP_COOKIE')
        res_type = 'LbCookiePersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type, {'id': LB_VS_ID},
            self._verify_create, 'default_cookie_name', 'INSERT')

    def test_setup_session_persistence_appcookie_new_profile(self):
        sess_persistence = lb_models.SessionPersistence(
            POOL_ID, 'APP_COOKIE', 'whatever')
        res_type = 'LbCookiePersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type, {'id': LB_VS_ID},
            self._verify_create, 'whatever', 'REWRITE')

    def test_setup_session_persistence_none_from_existing(self):
        sess_persistence = None
        self._test_setup_session_persistence(
            sess_persistence, None,
            {'id': LB_VS_ID, 'persistence_profile_id': LB_PP_ID},
            self._verify_delete)

    def test_setup_session_persistence_sourceip_from_existing(self):
        sess_persistence = lb_models.SessionPersistence(POOL_ID, 'SOURCE_IP')
        res_type = 'LbSourceIpPersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type,
            {'id': LB_VS_ID, 'persistence_profile_id': LB_PP_ID},
            self._verify_update)

    def test_setup_session_persistence_httpcookie_from_existing(self):
        sess_persistence = lb_models.SessionPersistence(POOL_ID, 'HTTP_COOKIE')
        res_type = 'LbCookiePersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type,
            {'id': LB_VS_ID, 'persistence_profile_id': LB_PP_ID},
            self._verify_update,
            'default_cookie_name', 'INSERT')

    def test_setup_session_persistence_appcookie_from_existing(self):
        sess_persistence = lb_models.SessionPersistence(
            POOL_ID, 'APP_COOKIE', 'whatever')
        res_type = 'LbCookiePersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type,
            {'id': LB_VS_ID, 'persistence_profile_id': LB_PP_ID},
            self._verify_update,
            'whatever', 'REWRITE')

    def test_setup_session_persistence_appcookie_switch_type(self):
        sess_persistence = lb_models.SessionPersistence(
            POOL_ID, 'APP_COOKIE', 'whatever')
        res_type = 'LbCookiePersistenceProfile'
        self._test_setup_session_persistence(
            sess_persistence, res_type,
            {'id': LB_VS_ID, 'persistence_profile_id': LB_PP_ID},
            self._verify_switch,
            'whatever', 'REWRITE',
            switch_type=True)


class TestEdgeLbaasV2Member(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2Member, self).setUp()

    @property
    def _tested_entity(self):
        return 'member'

    def test_create(self):
        with mock.patch.object(lb_utils, 'validate_lb_member_subnet'
                               ) as mock_validate_lb_subnet, \
            mock.patch.object(self.lbv2_driver.plugin, 'get_pool_members'
                              ) as mock_get_pool_members, \
            mock.patch.object(lb_utils, 'get_network_from_subnet'
                              ) as mock_get_network, \
            mock.patch.object(lb_utils, 'get_router_from_network'
                              ) as mock_get_router, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding'
                              ) as mock_get_lb_binding, \
            mock.patch.object(nsx_db, 'get_nsx_router_id'
                              ) as mock_get_nsx_router_id, \
            mock.patch.object(self.service_client, 'get_router_lb_service'
                              ) as mock_get_lb_service, \
            mock.patch.object(self.pool_client, 'get'
                              ) as mock_get_pool, \
            mock.patch.object(self.pool_client, 'update_pool_with_members'
                              ) as mock_update_pool_with_members:
            mock_validate_lb_subnet.return_value = True
            mock_get_pool_members.return_value = [self.member]
            mock_get_network.return_value = LB_NETWORK
            mock_get_router.return_value = LB_ROUTER_ID
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_lb_binding.return_value = LB_BINDING
            mock_get_nsx_router_id.return_value = LB_ROUTER_ID
            mock_get_lb_service.return_value = {'id': LB_SERVICE_ID}
            mock_get_pool.return_value = LB_POOL

            self.edge_driver.member.create(self.context, self.member)
            mock_update_pool_with_members.assert_called_with(LB_POOL_ID,
                                                             [LB_MEMBER])
            mock_successful_completion = (
                self.lbv2_driver.member.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.member,
                                                          delete=False)

    def test_create_external_vip(self):
        with mock.patch.object(lb_utils, 'validate_lb_member_subnet'
                               ) as mock_validate_lb_subnet, \
            mock.patch.object(self.lbv2_driver.plugin, 'get_pool_members'
                              ) as mock_get_pool_members, \
            mock.patch.object(lb_utils, 'get_network_from_subnet'
                              ) as mock_get_network, \
            mock.patch.object(lb_utils, 'get_router_from_network'
                              ) as mock_get_router, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_loadbalancer_binding',
                              ) as mock_get_lb_binding, \
            mock.patch.object(nsx_db, 'update_nsx_lbaas_loadbalancer_binding',
                              ) as mock_update_lb_binding, \
            mock.patch.object(nsx_db, 'get_nsx_router_id'
                              ) as mock_get_nsx_router_id, \
            mock.patch.object(self.service_client, 'get_router_lb_service'
                              ) as mock_get_lb_service, \
            mock.patch.object(self.pool_client, 'get'
                              ) as mock_get_pool, \
            mock.patch.object(self.core_plugin, '_find_router_gw_subnets',
                              return_value=[]),\
            mock.patch.object(self.pool_client, 'update_pool_with_members'
                              ) as mock_update_pool_with_members:
            mock_validate_lb_subnet.return_value = True
            mock_get_pool_members.return_value = [self.member]
            mock_get_network.return_value = LB_NETWORK
            mock_get_router.return_value = LB_ROUTER_ID
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_lb_binding.return_value = LB_BINDING_NO_RTR
            mock_get_nsx_router_id.return_value = LB_ROUTER_ID
            mock_get_lb_service.return_value = {'id': LB_SERVICE_ID}
            mock_get_pool.return_value = LB_POOL

            self.edge_driver.member.create(self.context, self.member)
            mock_update_pool_with_members.assert_called_with(LB_POOL_ID,
                                                             [LB_MEMBER])
            mock_successful_completion = (
                self.lbv2_driver.member.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.member,
                                                          delete=False)
            mock_update_lb_binding.assert_called_once_with(
                mock.ANY, LB_ID, LB_ROUTER_ID)

    def test_create_member_different_router(self):
        with mock.patch.object(self.lbv2_driver.plugin, 'get_pool_members'
                               ) as mock_get_pool_members, \
            mock.patch.object(lb_utils, 'get_network_from_subnet'
                              ) as mock_get_network, \
            mock.patch.object(lb_utils, 'get_router_from_network'
                              ) as mock_get_router:
            mock_get_pool_members.return_value = [self.member]
            mock_get_network.return_value = LB_NETWORK
            mock_get_router.return_value = None

            self.assertRaises(n_exc.BadRequest,
                              self.edge_driver.member.create,
                              self.context,
                              self.member)

    def test_update(self):
        new_member = lb_models.Member(MEMBER_ID, LB_TENANT_ID, POOL_ID,
                                      MEMBER_ADDRESS, 80, 2, pool=self.pool,
                                      name='member-nnn-nnn')
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding, \
            mock.patch.object(self.pool_client, 'get'
                              ) as mock_get_pool, \
            mock.patch.object(lb_utils, 'get_network_from_subnet'
                              ) as mock_get_network_from_subnet:
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_pool.return_value = LB_POOL_WITH_MEMBER
            mock_get_network_from_subnet.return_value = LB_NETWORK

            self.edge_driver.member.update(self.context, self.member,
                                           new_member)

            mock_successful_completion = (
                self.lbv2_driver.member.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_member,
                                                          delete=False)

    def test_delete(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                               ) as mock_get_pool_binding, \
            mock.patch.object(self.pool_client, 'get'
                              ) as mock_get_pool, \
            mock.patch.object(lb_utils, 'get_network_from_subnet'
                              ) as mock_get_network_from_subnet, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(self.pool_client, 'update_pool_with_members'
                              ) as mock_update_pool_with_members:
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_pool.return_value = LB_POOL_WITH_MEMBER
            mock_get_network_from_subnet.return_value = LB_NETWORK
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID

            self.edge_driver.member.delete(self.context, self.member)

            mock_update_pool_with_members.assert_called_with(LB_POOL_ID, [])

            mock_successful_completion = (
                self.lbv2_driver.member.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.member,
                                                          delete=True)


class TestEdgeLbaasV2HealthMonitor(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2HealthMonitor, self).setUp()

    @property
    def _tested_entity(self):
        return 'health_monitor'

    def test_create(self):
        with mock.patch.object(self.monitor_client, 'create'
                               ) as mock_create_monitor, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.pool_client, 'add_monitor_to_pool'
                              ) as mock_add_monitor_to_pool, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_monitor_binding'
                              ) as mock_add_monitor_binding:
            mock_create_monitor.return_value = {'id': LB_MONITOR_ID}
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.healthmonitor.create(self.context, self.hm)

            mock_add_monitor_to_pool.assert_called_with(LB_POOL_ID,
                                                        LB_MONITOR_ID)
            mock_add_monitor_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, HM_ID, LB_MONITOR_ID,
                LB_POOL_ID)

            mock_successful_completion = (
                self.lbv2_driver.health_monitor.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.hm,
                                                          delete=False)

    def test_create_http(self):
        with mock.patch.object(self.monitor_client, 'create'
                               ) as mock_create_monitor, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.pool_client, 'add_monitor_to_pool'
                              ) as mock_add_monitor_to_pool, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_monitor_binding'
                              ) as mock_add_monitor_binding:
            mock_create_monitor.return_value = {'id': LB_MONITOR_ID}
            mock_get_pool_binding.return_value = POOL_BINDING

            # Verify HTTP-specific  monitor parameters are added
            self.edge_driver.healthmonitor.create(self.context, self.hm_http)
            self.assertEqual(1, len(mock_create_monitor.mock_calls))
            kw_args = mock_create_monitor.mock_calls[0][2]
            self.assertEqual(self.hm_http.http_method,
                             kw_args.get('request_method'))
            self.assertEqual(self.hm_http.url_path,
                             kw_args.get('request_url'))
            mock_add_monitor_to_pool.assert_called_with(LB_POOL_ID,
                                                        LB_MONITOR_ID)
            mock_add_monitor_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, HM_ID, LB_MONITOR_ID,
                LB_POOL_ID)

            mock_successful_completion = (
                self.lbv2_driver.health_monitor.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.hm_http,
                                                          delete=False)

    def test_update(self):
        with mock.patch.object(self.monitor_client, 'update'
                               ) as mock_update_monitor, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_monitor_binding'
                              ) as mock_get_monitor_binding:
            mock_get_monitor_binding.return_value = HM_BINDING
            new_hm = lb_models.HealthMonitor(
                HM_ID, LB_TENANT_ID, 'PING', 5, 5,
                5, pool=self.pool, name='new_name')
            self.edge_driver.healthmonitor.update(
                self.context, self.hm, new_hm)
            mock_update_monitor.assert_called_with(
                LB_MONITOR_ID, display_name=mock.ANY,
                fall_count=5, interval=5, timeout=5,
                resource_type='LbIcmpMonitor')

            mock_successful_completion = (
                self.lbv2_driver.health_monitor.successful_completion)
            mock_successful_completion.assert_called_with(self.context, new_hm,
                                                          delete=False)

    def test_delete(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_monitor_binding'
                               ) as mock_get_monitor_binding, \
            mock.patch.object(self.pool_client, 'remove_monitor_from_pool'
                              ) as mock_remove_monitor_from_pool, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(self.monitor_client, 'delete'
                              ) as mock_delete_monitor, \
            mock.patch.object(nsx_db, 'delete_nsx_lbaas_monitor_binding'
                              ) as mock_delete_monitor_binding:
            mock_get_monitor_binding.return_value = HM_BINDING

            self.edge_driver.healthmonitor.delete(self.context, self.hm)

            mock_remove_monitor_from_pool.assert_called_with(LB_POOL_ID,
                                                             LB_MONITOR_ID)
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID
            mock_delete_monitor.assert_called_with(LB_MONITOR_ID)
            mock_delete_monitor_binding.assert_called_with(
                self.context.session, LB_ID, POOL_ID, HM_ID)

            mock_successful_completion = (
                self.lbv2_driver.health_monitor.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.hm,
                                                          delete=True)


class TestEdgeLbaasV2L7Policy(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2L7Policy, self).setUp()

    @property
    def _tested_entity(self):
        return 'l7policy'

    def test_create(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_listener_binding'
                               ) as mock_get_listener_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.rule_client, 'create'
                              ) as mock_create_rule, \
            mock.patch.object(self.vs_client, 'get'
                              ) as mock_get_virtual_server, \
            mock.patch.object(self.vs_client, 'update'
                              ) as mock_update_virtual_server, \
            mock.patch.object(nsx_db, 'add_nsx_lbaas_l7policy_binding'
                              ) as mock_add_l7policy_binding:
            mock_get_listener_binding.return_value = LISTENER_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_create_rule.return_value = {'id': LB_RULE_ID}
            mock_get_virtual_server.return_value = {'id': LB_VS_ID}

            self.edge_driver.l7policy.create(self.context, self.l7policy)

            mock_update_virtual_server.assert_called_with(
                LB_VS_ID, rule_ids=[LB_RULE_ID])
            mock_add_l7policy_binding.assert_called_with(
                self.context.session, L7POLICY_ID, LB_RULE_ID, LB_VS_ID)
            mock_successful_completion = (
                self.lbv2_driver.l7policy.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.l7policy,
                                                          delete=False)

    def test_update(self):
        new_l7policy = lb_models.L7Policy(L7POLICY_ID, LB_TENANT_ID,
                                          name='new-policy',
                                          listener_id=LISTENER_ID,
                                          action='REJECT',
                                          listener=self.listener,
                                          position=2)
        vs_with_rules = {
            'id': LB_VS_ID,
            'rule_ids': [LB_RULE_ID, 'abc', 'xyz']
        }
        rule_body = {
            'match_conditions': [],
            'actions': [{
                'type': 'LbHttpRejectAction',
                'reply_status': '403'}],
            'phase': 'HTTP_FORWARDING',
            'match_strategy': 'ALL'
        }
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_l7policy_binding'
                               ) as mock_get_l7policy_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.rule_client, 'update'
                              ) as mock_update_rule, \
            mock.patch.object(self.vs_client, 'get'
                              ) as mock_get_virtual_server, \
            mock.patch.object(self.vs_client, 'update'
                              ) as mock_update_virtual_server:
            mock_get_l7policy_binding.return_value = L7POLICY_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING
            mock_get_virtual_server.return_value = vs_with_rules

            self.edge_driver.l7policy.update(self.context, self.l7policy,
                                             new_l7policy)

            mock_update_rule.assert_called_with(LB_RULE_ID,
                                                **rule_body)
            mock_update_virtual_server.assert_called_with(
                LB_VS_ID, rule_ids=['abc', LB_RULE_ID, 'xyz'])
            mock_successful_completion = (
                self.lbv2_driver.l7policy.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_l7policy,
                                                          delete=False)

    def test_delete(self):
        with mock.patch.object(nsx_db, 'get_nsx_lbaas_l7policy_binding'
                               ) as mock_get_l7policy_binding, \
            mock.patch.object(self.vs_client, 'remove_rule'
                              ) as mock_vs_remove_rule, \
            mock.patch.object(self.rule_client, 'delete'
                              ) as mock_delete_rule, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(nsx_db, 'delete_nsx_lbaas_l7policy_binding'
                              ) as mock_delete_l7policy_binding:
            mock_get_l7policy_binding.return_value = L7POLICY_BINDING
            mock_get_neutron_from_nsx_router_id.return_value = LB_ROUTER_ID

            self.edge_driver.l7policy.delete(self.context, self.l7policy)

            mock_vs_remove_rule.assert_called_with(LB_VS_ID, LB_RULE_ID)
            mock_delete_rule.assert_called_with(LB_RULE_ID)
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID
            mock_delete_l7policy_binding.assert_called_with(
                self.context.session, L7POLICY_ID)
            mock_successful_completion = (
                self.lbv2_driver.l7policy.successful_completion)
            mock_successful_completion.assert_called_with(
                self.context, self.l7policy, delete=True)


class TestEdgeLbaasV2L7Rule(BaseTestEdgeLbaasV2):
    def setUp(self):
        super(TestEdgeLbaasV2L7Rule, self).setUp()

    @property
    def _tested_entity(self):
        return 'l7rule'

    def test_create(self):
        self.l7policy.rules = [self.l7rule]
        create_rule_body = {
            'match_conditions': [{
                'type': 'LbHttpRequestHeaderCondition',
                'match_type': 'EQUALS',
                'header_name': self.l7rule.key,
                'header_value': self.l7rule.value}],
            'actions': [{
                'type': 'LbSelectPoolAction',
                'pool_id': LB_POOL_ID}],
            'phase': 'HTTP_FORWARDING',
            'match_strategy': 'ALL'
        }

        with mock.patch.object(nsx_db, 'get_nsx_lbaas_l7policy_binding'
                               ) as mock_get_l7policy_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.rule_client, 'update'
                              ) as mock_update_rule:
            mock_get_l7policy_binding.return_value = L7POLICY_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.l7rule.create(self.context, self.l7rule)

            mock_update_rule.assert_called_with(LB_RULE_ID,
                                                **create_rule_body)

            mock_successful_completion = (
                self.lbv2_driver.l7rule.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.l7rule,
                                                          delete=False)

    def test_update(self):
        new_l7rule = lb_models.L7Rule(L7RULE_ID, LB_TENANT_ID,
                                      l7policy_id=L7POLICY_ID,
                                      compare_type='STARTS_WITH',
                                      invert=True,
                                      type='COOKIE',
                                      key='cookie1',
                                      value='xxxxx',
                                      policy=self.l7policy)
        self.l7policy.rules = [new_l7rule]
        update_rule_body = {
            'match_conditions': [{
                'type': 'LbHttpRequestHeaderCondition',
                'match_type': 'STARTS_WITH',
                'header_name': 'Cookie',
                'header_value': 'cookie1=xxxxx'}],
            'actions': [{
                'type': 'LbSelectPoolAction',
                'pool_id': LB_POOL_ID}],
            'phase': 'HTTP_FORWARDING',
            'match_strategy': 'ALL'
        }

        with mock.patch.object(nsx_db, 'get_nsx_lbaas_l7policy_binding'
                               ) as mock_get_l7policy_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(self.rule_client, 'update'
                              ) as mock_update_rule:
            mock_get_l7policy_binding.return_value = L7POLICY_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.l7rule.update(self.context, self.l7rule,
                                           new_l7rule)

            mock_update_rule.assert_called_with(LB_RULE_ID,
                                                **update_rule_body)

            mock_successful_completion = (
                self.lbv2_driver.l7rule.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          new_l7rule,
                                                          delete=False)

    def test_delete(self):
        self.l7policy.rules = [self.l7rule]
        delete_rule_body = {
            'match_conditions': [],
            'actions': [{
                'type': 'LbSelectPoolAction',
                'pool_id': LB_POOL_ID}],
            'phase': 'HTTP_FORWARDING',
            'match_strategy': 'ALL'
        }

        with mock.patch.object(nsx_db, 'get_nsx_lbaas_l7policy_binding'
                               ) as mock_get_l7policy_binding, \
            mock.patch.object(nsx_db, 'get_nsx_lbaas_pool_binding'
                              ) as mock_get_pool_binding, \
            mock.patch.object(nsx_db, 'get_neutron_from_nsx_router_id'
                              ) as mock_get_neutron_from_nsx_router_id, \
            mock.patch.object(self.rule_client, 'update'
                              ) as mock_update_rule:
            mock_get_l7policy_binding.return_value = L7POLICY_BINDING
            mock_get_pool_binding.return_value = POOL_BINDING

            self.edge_driver.l7rule.delete(self.context, self.l7rule)

            mock_update_rule.assert_called_with(LB_RULE_ID,
                                                **delete_rule_body)
            mock_get_neutron_from_nsx_router_id.router_id = ROUTER_ID

            mock_successful_completion = (
                self.lbv2_driver.l7rule.successful_completion)
            mock_successful_completion.assert_called_with(self.context,
                                                          self.l7rule,
                                                          delete=True)



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""NSXv add dns search domain to subnets

Revision ID: 3e4dccfe6fb4
Revises: 2c87aedb206f
Create Date: 2016-03-20 07:28:35.369938

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3e4dccfe6fb4'
down_revision = '2c87aedb206f'


def upgrade():
    op.create_table(
        'nsx_extended_security_group_properties',
        sa.Column('security_group_id', sa.String(36), nullable=False),
        sa.Column('logging', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['security_group_id'],
                                ['securitygroups.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('security_group_id')
    )



# Copyright 2018 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""lbaas_no_foreign_key

Revision ID: fc6308289aca
Revises: 0dbeda408e41
Create Date: 2018-06-04 13:47:09.450116
"""

from alembic import op
from sqlalchemy.engine import reflection

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = 'fc6308289aca'
down_revision = '0dbeda408e41'
depends_on = ('717f7f63a219')


def upgrade():
    for table_name in ['nsxv3_lbaas_loadbalancers',
                       'nsxv3_lbaas_listeners',
                       'nsxv3_lbaas_pools',
                       'nsxv3_lbaas_monitors',
                       'nsxv3_lbaas_l7rules',
                       'nsxv3_lbaas_l7policies',
                       'nsxv_lbaas_loadbalancer_bindings',
                       'nsxv_lbaas_listener_bindings',
                       'nsxv_lbaas_pool_bindings',
                       'nsxv_lbaas_monitor_bindings',
                       'nsxv_lbaas_l7policy_bindings']:

        if migration.schema_has_table(table_name):
            inspector = reflection.Inspector.from_engine(op.get_bind())
            fk_constraint = inspector.get_foreign_keys(table_name)[0]
            op.drop_constraint(fk_constraint.get('name'), table_name,
                               type_='foreignkey')
